from pathlib import Path

from duckdb import BinderException
import pandas as pd
import pytest

from viadot.sources import DuckDB


TABLE = "test_table"
SCHEMA = "test_schema"
TABLE_MULTIPLE_PARQUETS = "test_multiple_parquets"
DATABASE_PATH = "test_db_123.duckdb"


@pytest.fixture(scope="module")
def duckdb():
    duckdb = DuckDB(credentials={"database": DATABASE_PATH, "read_only": False})
    yield duckdb
    Path(DATABASE_PATH).unlink()


def test_create_table_from_parquet(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )
    df = duckdb.to_df(f"SELECT * FROM {SCHEMA}.{TABLE}")  # noqa: S608
    assert df.shape[0] == 3
    duckdb.drop_table(TABLE, schema=SCHEMA)
    duckdb.run_query(f"DROP SCHEMA {SCHEMA}")


def test_create_table_from_parquet_append(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )
    df = duckdb.to_df(f"SELECT * FROM {SCHEMA}.{TABLE}")  # noqa: S608
    assert df.shape[0] == 3

    # now append
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH, if_exists="append"
    )
    df = duckdb.to_df(f"SELECT * FROM {SCHEMA}.{TABLE}")  # noqa: S608
    assert df.shape[0] == 6

    duckdb.drop_table(TABLE, schema=SCHEMA)
    duckdb.run_query(f"DROP SCHEMA {SCHEMA}")


def test_create_table_from_parquet_delete(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )
    df = duckdb.to_df(f"SELECT * FROM {SCHEMA}.{TABLE}")  # noqa: S608
    assert df.shape[0] == 3

    df = pd.DataFrame.from_dict(
        data={
            "country": ["italy", "germany", "spain"],
            "sales": [100, 50, 80],
            "color": ["red", "blue", "grren"],
        }
    )

    df.to_parquet("test_parquet.parquet")
    with pytest.raises(BinderException):
        duckdb.create_table_from_parquet(
            schema=SCHEMA, table=TABLE, path="test_parquet.parquet", if_exists="delete"
        )

    duckdb.drop_table(TABLE, schema=SCHEMA)
    duckdb.run_query(f"DROP SCHEMA {SCHEMA}")
    Path("test_parquet.parquet").unlink()


def test_create_table_from_multiple_parquet(duckdb):
    # we use the two Parquet files generated by fixtures in conftest
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE_MULTIPLE_PARQUETS, path="*.parquet"
    )
    df = duckdb.to_df(f"SELECT * FROM {SCHEMA}.{TABLE_MULTIPLE_PARQUETS}")  # noqa: S608
    assert df.shape[0] == 6
    duckdb.drop_table(TABLE_MULTIPLE_PARQUETS, schema=SCHEMA)
    duckdb.run_query(f"DROP SCHEMA {SCHEMA}")


def test_check_if_table_exists(duckdb, TEST_PARQUET_FILE_PATH):
    assert not duckdb._check_if_table_exists(table=TABLE, schema=SCHEMA)
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )
    assert duckdb._check_if_table_exists(TABLE, schema=SCHEMA)
    duckdb.drop_table(TABLE, schema=SCHEMA)


def test_run_query_query_with_comments(duckdb, TEST_PARQUET_FILE_PATH):
    duckdb.create_table_from_parquet(
        schema=SCHEMA, table=TABLE, path=TEST_PARQUET_FILE_PATH
    )
    output1 = duckdb.run_query(
        query=f"""
        --test
    SELECT * FROM {SCHEMA}.{TABLE}
    """,  # noqa: S608
        fetch_type="dataframe",
    )
    assert isinstance(output1, pd.DataFrame)

    output2 = duckdb.run_query(
        query=f"""
    SELECT * FROM {SCHEMA}.{TABLE}
    WHERE country = 'italy'
    """,  # noqa: S608
        fetch_type="dataframe",
    )
    assert isinstance(output2, pd.DataFrame)

    output3 = duckdb.run_query(
        query=f"""
    SELECT * FROM {SCHEMA}.{TABLE}
        ---test
    """,  # noqa: S608
        fetch_type="dataframe",
    )
    assert isinstance(output3, pd.DataFrame)

    duckdb.drop_table(TABLE, schema=SCHEMA)



"""Flows for downloading data from SAP and uploading it to AWS Redshift Spectrum."""

from typing import Any, Literal

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_redshift_spectrum, sap_rfc_to_df


@flow(
    name="extract--sap--redshift_spectrum",
    description="Extract data from SAP and load it into AWS Redshift Spectrum.",
    retries=1,
    retry_delay_seconds=60,
)
def sap_to_redshift_spectrum(  # noqa: PLR0913
    to_path: str,
    schema_name: str,
    table: str,
    tests: dict[str, Any] | None = None,
    extension: str = ".parquet",
    if_exists: Literal["overwrite", "append"] = "overwrite",
    partition_cols: list[str] | None = None,
    index: bool = False,
    compression: str | None = None,
    aws_sep: str = ",",
    dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
    dynamic_date_format: str = "%Y%m%d",
    dynamic_date_timezone: str = "UTC",
    credentials_secret: str | None = None,
    aws_config_key: str | None = None,
    query: str | None = None,
    func: str | None = None,
    rfc_total_col_width_character_limit: int = 400,
    rfc_unique_id: list[str] | None = None,
    sap_credentials_secret: str | None = None,
    sap_config_key: str | None = None,
    sap_sep: str | None = "♔",
    replacement: str = "-",
) -> None:
    """Download a pandas `DataFrame` from SAP and upload it to AWS Redshift Spectrum.

    Args:
        to_path (str): Path to a S3 folder where the table will be located.
            Defaults to None.
        schema_name (str): AWS Glue catalog database name.
        table (str): AWS Glue catalog table name.
        tests (dict[str], optional): A dictionary with optional list of tests
            to verify the output dataframe. If defined, triggers the `validate`
            function from viadot.utils. Defaults to None.
        partition_cols (list[str]): List of column names that will be used to create
            partitions. Only takes effect if dataset=True.
        extension (str): Required file type. Accepted file formats are 'csv' and
            'parquet'.
        if_exists (str, optional): 'overwrite' to recreate any possible existing table
            or 'append' to keep any possible existing table. Defaults to overwrite.
        partition_cols (list[str], optional): List of column names that will be used to
            create partitions. Only takes effect if dataset=True. Defaults to None.
        index (bool, optional): Write row names (index). Defaults to False.
        compression (str, optional): Compression style (None, snappy, gzip, zstd).
        aws_sep (str, optional): Field delimiter for the output file. Defaults to ','.
        dynamic_date_symbols (list[str], optional): Symbols used for dynamic date
            handling. Defaults to ["<<", ">>"].
        dynamic_date_format (str, optional): Format used for dynamic date parsing.
            Defaults to "%Y%m%d".
        dynamic_date_timezone (str, optional): Timezone used for dynamic date
            processing. Defaults to "UTC".
        aws_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        credentials_secret (str, optional): The name of a secret block in Prefect
            that stores AWS credentials. Defaults to None.
        query (str): The query to be executed with pyRFC.
        func (str, optional): SAP RFC function to use. Defaults to None.
        rfc_total_col_width_character_limit (int, optional): Number of characters by
            which query will be split in chunks in case of too many columns for RFC
            function. According to SAP documentation, the limit is 512 characters.
            However, we observed SAP raising an exception even on a slightly lower
            number of characters, so we add a safety margin. Defaults to 400.
        rfc_unique_id  (list[str], optional): Reference columns to merge chunks Data
            Frames. These columns must to be unique. If no columns are provided, all
                data frame columns will by concatenated. Defaults to None.
        sap_credentials_secret (str, optional): The name of the AWS secret that stores
            SAP credentials. Defaults to None.
        sap_config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to None.
        replacement (str, optional): In case of sep is on a columns, set up a new
            character to replace inside the string to avoid flow breakdowns.
            Defaults to "-".
        sap_sep (str, optional): The separator to use when reading query results.
            If set to None, multiple options are automatically tried.
            Defaults to ♔.


    Examples:
        sap_to_redshift_spectrum(
            ...
            rfc_unique_id=["VBELN", "LPRIO"],
            ...
        )
    """
    df = sap_rfc_to_df(
        query=query,
        tests=tests,
        func=func,
        rfc_unique_id=rfc_unique_id,
        rfc_total_col_width_character_limit=rfc_total_col_width_character_limit,
        credentials_secret=sap_credentials_secret,
        config_key=sap_config_key,
        dynamic_date_symbols=dynamic_date_symbols,
        dynamic_date_format=dynamic_date_format,
        dynamic_date_timezone=dynamic_date_timezone,
        replacement=replacement,
        sep=sap_sep,
    )

    return df_to_redshift_spectrum(
        df=df,
        to_path=to_path,
        schema_name=schema_name,
        table=table,
        extension=extension,
        if_exists=if_exists,
        partition_cols=partition_cols,
        index=index,
        compression=compression,
        sep=aws_sep,
        config_key=aws_config_key,
        credentials_secret=credentials_secret,
    )



"""Task to download data from Outlook API into a Pandas DataFrame."""

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Outlook


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def outlook_to_df(
    mailbox_name: str,
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    request_retries: int = 10,
    start_date: str | None = None,
    end_date: str | None = None,
    limit: int = 10000,
    address_limit: int = 8000,
    outbox_list: list[str] | None = None,
) -> pd.DataFrame:
    """Task for downloading data from Outlook API to a pandas DataFrame.

    Args:
        mailbox_name (str): Mailbox name.
        config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        request_retries (int, optional): How many times retries to authorizate.
            Defaults to 10.
        start_date (Optional[str], optional): A filtering start date parameter e.g.
            "2022-01-01". Defaults to None.
        end_date (Optional[str], optional): A filtering end date parameter e.g.
            "2022-01-02". Defaults to None.
        limit (int, optional): Number of fetched top messages. Defaults to 10000.
        address_limit (int, optional): The maximum number of accepted characters in the
            sum of all email names. Defaults to 8000.
        outbox_list (List[str], optional): List of outbox folders to differentiate
            between Inboxes and Outboxes. Defaults to ["Sent Items"].

    Examples:
        data_frame = mindful_to_df(
            config_key=config_key,
            azure_key_vault_secret=azure_key_vault_secret,
            region=region,
            endpoint=end,
            date_interval=date_interval,
            limit=limit,
        )

    Raises:
        MissingSourceCredentialsError: If none credentials have been provided.
        APIError: The mailbox name is a "must" requirement.

    Returns:
        pd.DataFrame: The response data as a pandas DataFrame.
    """
    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    if not outbox_list:
        outbox_list = ["Sent Items"]

    outlook = Outlook(
        credentials=credentials,
        config_key=config_key,
    )
    outlook.api_connection(
        mailbox_name=mailbox_name,
        request_retries=request_retries,
        start_date=start_date,
        end_date=end_date,
        limit=limit,
        address_limit=address_limit,
        outbox_list=outbox_list,
    )

    return outlook.to_df()



from contextlib import nullcontext as does_not_raise
import json
import logging
import re

import pandas as pd
import pytest

from viadot.signals import SKIP
from viadot.utils import (
    _cast_df_cols,
    add_viadot_metadata_columns,
    df_clean_column,
    df_converts_bytes_to_int,
    gen_bulk_insert_query_from_df,
    get_fqn,
    handle_api_request,
    handle_if_empty,
    validate,
    validate_and_reorder_dfs_columns,
)


def test_single_quotes_inside():
    test_value = "a'b"
    df1 = pd.DataFrame(
        {
            "a": [
                test_value,
            ],
            "b": ["a"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    test_value_escaped = "'a''b'"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({test_value_escaped}, 'a')"""
    ), test_insert_query


def test_single_quotes_outside():
    test_value = "'a'"
    df1 = pd.DataFrame(
        {
            "a": [
                test_value,
            ],
            "b": ["b"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    test_value_escaped = "'''a'''"
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({test_value_escaped}, 'b')"""
    ), test_insert_query


def test_double_quotes_inside():
    test_value = 'a "b"'
    df1 = pd.DataFrame(
        {
            "a": [
                test_value,
            ],
            "b": ["c"],
        }
    )
    test_insert_query = gen_bulk_insert_query_from_df(
        df1, table_fqn="test_schema.test_table"
    )
    test_value_escaped = """'a "b"'"""
    assert (
        test_insert_query
        == f"""INSERT INTO test_schema.test_table (a, b)

VALUES ({test_value_escaped}, 'c')"""
    ), test_insert_query


def test_handle_api_request():
    url = "https://api.restful-api.dev/objects"
    headers = {"content-type": "application/json"}
    item = {
        "name": "test_item",
        "data": {"color": "blue", "price": 135},
    }
    payload = json.dumps(item)

    response_post = handle_api_request(
        url=url, method="POST", headers=headers, data=payload
    )
    assert response_post.ok

    item_url = f"""{url}/{response_post.json()["id"]}"""
    response_get = handle_api_request(url=item_url, method="GET", headers=headers)
    assert response_get.ok
    assert response_get.json()["data"] == item["data"]

    response_delete = handle_api_request(url=item_url, method="DELETE", headers=headers)
    assert response_delete.ok


def test_add_viadot_metadata_columns():
    class TestingClass:
        @add_viadot_metadata_columns
        def to_df(self):
            my_dict = {"AA": [1, 1], "BB": [2, 2]}
            return pd.DataFrame(my_dict)

    testing_instance = TestingClass()
    df = testing_instance.to_df()
    assert "_viadot_source" in df.columns


def test__cast_df_cols():
    test_df = pd.DataFrame(
        {
            "bool_column": [True, False, True, False],
            "datetime_column": [
                "2023-05-25 10:30:00",
                "2023-05-20 10:00:00",
                "2023-05-15 10:30:00",
                "2023-05-10 10:30:00",
            ],
            "int_column": [5, 10, 15, 20],
            "object_column": ["apple", "banana", "melon", "orange"],
        }
    )
    test_df["datetime_column"] = pd.to_datetime(
        test_df["datetime_column"], infer_datetime_format=True
    )
    result_df = _cast_df_cols(
        test_df, types_to_convert=["datetime", "bool", "int", "object"]
    )

    assert result_df["bool_column"].dtype == pd.Int64Dtype()
    assert pd.api.types.is_object_dtype(result_df["datetime_column"])
    assert result_df["int_column"].dtype == pd.Int64Dtype()
    assert result_df["object_column"].dtype == pd.StringDtype()


def test_get_fqn():
    # Test with schema name.
    fqn = get_fqn(table_name="my_table", schema_name="my_schema")
    assert fqn == "my_schema.my_table"

    # Test without schema name.
    fqn = get_fqn(table_name="my_table")
    assert fqn == "my_table"


def test_validate_column_size_pass():
    df = pd.DataFrame({"col1": ["a", "bb", "ccc"]})
    tests = {"column_size": {"col1": 3}}
    with does_not_raise():
        validate(df, tests)


def test_validate_column_size_fail(caplog):
    df = pd.DataFrame({"col1": ["a", "bb", "cccc"]})
    tests = {"column_size": {"col1": 3}}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception, match=r"Validation failed for 1 test\(s\): column_size error"
        ),
    ):
        validate(df, tests)

    assert "field length is different than 3" in caplog.text


def test_validate_column_unique_values_pass():
    df = pd.DataFrame({"col1": [1, 2, 3]})
    tests = {"column_unique_values": ["col1"]}
    with does_not_raise():
        validate(df, tests)


def test_validate_column_unique_values_fail(caplog):
    df = pd.DataFrame({"col1": [1, 2, 2]})
    tests = {"column_unique_values": ["col1"]}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception,
            match=r"Validation failed for 1 test\(s\): column_unique_values error",
        ),
    ):
        validate(df, tests)

    assert "Values for col1 are not unique." in caplog.text


def test_validate_column_list_to_match_pass():
    df = pd.DataFrame({"col1": [1], "col2": [2]})
    tests = {"column_list_to_match": ["col1", "col2"]}
    with does_not_raise():
        validate(df, tests)


def test_validate_column_list_to_match_fail(caplog):
    df = pd.DataFrame({"col1": [1]})
    tests = {"column_list_to_match": ["col1", "col2"]}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception,
            match=r"Validation failed for 1 test\(s\): column_list_to_match error",
        ),
    ):
        validate(df, tests)

    assert "Columns are different than expected" in caplog.text


def test_validate_dataset_row_count_pass():
    df = pd.DataFrame({"col1": [1, 2, 3]})
    tests = {"dataset_row_count": {"min": 1, "max": 5}}
    with does_not_raise():
        validate(df, tests)


def test_validate_dataset_row_count_fail(caplog):
    df = pd.DataFrame({"col1": [1, 2, 3, 4, 5, 6]})
    tests = {"dataset_row_count": {"min": 1, "max": 5}}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception,
            match=r"Validation failed for 1 test\(s\): dataset_row_count error",
        ),
    ):
        validate(df, tests)

    assert "Row count (6) is not between 1 and 5" in caplog.text


def test_validate_column_match_regex_pass():
    df = pd.DataFrame({"col1": ["A12", "B34", "C45"]})
    tests = {"column_match_regex": {"col1": "^[A-Z][0-9]{2}$"}}
    with does_not_raise():
        validate(df, tests)


def test_validate_column_match_regex_fail(caplog):
    df = pd.DataFrame({"col1": ["A123", "B34", "C45"]})
    tests = {"column_match_regex": {"col1": "^[A-Z][0-9]{2}$"}}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception,
            match=r"Validation failed for 1 test\(s\): column_match_regex error",
        ),
    ):
        validate(df, tests)

    assert "[column_match_regex] on col1 column failed!" in caplog.text


def test_validate_column_sum_pass():
    df = pd.DataFrame({"col1": [1, 2, 3]})
    tests = {"column_sum": {"col1": {"min": 5, "max": 10}}}
    with does_not_raise():
        validate(df, tests)


def test_validate_column_sum_fail(caplog):
    df = pd.DataFrame({"col1": [1, 2, 3, 4]})
    tests = {"column_sum": {"col1": {"min": 5, "max": 6}}}

    with (
        caplog.at_level(logging.INFO),
        pytest.raises(
            Exception, match=r"Validation failed for \d+ test\(s\): column_sum error"
        ),
    ):
        validate(df, tests)

    assert "Sum of 10 for col1 is out of the expected range - <5:6>" in caplog.text


def test_validate_and_reorder_wrong_columns():
    df1 = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    df2 = pd.DataFrame({"a": [5, 6], "c": [7, 8]})

    with pytest.raises(ValueError):  # noqa: PT011
        validate_and_reorder_dfs_columns([df1, df2])


def test_validate_and_reorder_empty_list():
    with pytest.raises(IndexError):
        validate_and_reorder_dfs_columns([])


def test_validate_and_reorder_identical_columns():
    df1 = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    df2 = pd.DataFrame({"a": [5, 6], "b": [7, 8]})

    result = validate_and_reorder_dfs_columns([df1, df2])

    assert len(result) == 2
    assert list(result[0].columns) == list(df1.columns)
    assert result[0].equals(df1)
    assert list(result[1].columns) == list(df2.columns)
    assert result[1].equals(df2)


def test_validate_and_reorder_different_order_columns():
    df1 = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
    df2 = pd.DataFrame({"b": [7, 8], "a": [5, 6]})

    expected_df2 = pd.DataFrame({"a": [5, 6], "b": [7, 8]})
    result = validate_and_reorder_dfs_columns([df1, df2])

    assert len(result) == 2
    assert list(result[0].columns) == list(df1.columns)
    assert result[0].equals(df1)
    assert list(result[1].columns) == list(expected_df2.columns)
    assert result[1].equals(expected_df2)


def test_df_converts_bytes_to_int():
    df_bytes = pd.DataFrame(
        {
            "A": [b"1", b"2", b"3"],
            "B": [b"4", b"5", b"6"],
            "C": ["no change", "still no change", "me neither"],
        }
    )

    result = df_converts_bytes_to_int(df_bytes)

    expected = pd.DataFrame(
        {
            "A": [1, 2, 3],
            "B": [4, 5, 6],
            "C": ["no change", "still no change", "me neither"],
        }
    )

    pd.testing.assert_frame_equal(result, expected)

    df_no_bytes = pd.DataFrame({"A": [1, 2, 3], "B": [4, 5, 6]})

    result_no_bytes = df_converts_bytes_to_int(df_no_bytes)

    pd.testing.assert_frame_equal(result_no_bytes, df_no_bytes)


def test_df_clean_column():
    df_dirty = pd.DataFrame(
        {
            "A": ["Hello\tWorld", "Goodbye\nWorld"],
            "B": ["Keep\nIt\tClean", "Just\tTest"],
        }
    )

    cleaned_df = df_clean_column(df_dirty, columns_to_clean=["A"])

    expected_cleaned_df = pd.DataFrame(
        {"A": ["HelloWorld", "GoodbyeWorld"], "B": ["Keep\nIt\tClean", "Just\tTest"]}
    )

    pd.testing.assert_frame_equal(cleaned_df, expected_cleaned_df)

    cleaned_all_df = df_clean_column(df_dirty)

    expected_all_cleaned_df = pd.DataFrame(
        {"A": ["HelloWorld", "GoodbyeWorld"], "B": ["KeepItClean", "JustTest"]}
    )

    pd.testing.assert_frame_equal(cleaned_all_df, expected_all_cleaned_df)


def test_handle_if_empty_warn(caplog):
    """Test that handle_if_empty logs a warning when if_empty='warn'."""
    message = "Empty input detected."
    caplog.set_level(logging.WARNING)

    handle_if_empty(if_empty="warn", message=message)

    assert message in caplog.text
    assert "WARNING" in caplog.text


def test_handle_if_empty_skip():
    """Test that handle_if_empty raises SKIP exception when if_empty='skip'."""
    message = "Skipping due to empty input."

    with pytest.raises(SKIP, match=message):
        handle_if_empty(if_empty="skip", message=message)


def test_handle_if_empty_fail():
    """Test that handle_if_empty raises ValueError when if_empty='fail'."""
    message = "Failed due to empty input."

    with pytest.raises(ValueError, match=message):
        handle_if_empty(if_empty="fail", message=message)


def test_handle_if_empty_default_logger(caplog):
    """Test that handle_if_empty uses default logger when none is provided."""
    message = "Empty input detected with default logger."
    caplog.set_level(logging.WARNING)

    handle_if_empty(if_empty="warn", message=message)

    assert message in caplog.text
    assert "WARNING" in caplog.text


def test_handle_if_empty_invalid_value():
    """Test that handle_if_empty raises ValueError when an invalid if_empty value is provided."""  # noqa: W505
    invalid_value = "replace"
    expected_msg = f"Invalid value for if_empty: {invalid_value}. Allowed values are ['warn', 'skip', 'fail']."
    escaped_msg = re.escape(expected_msg)  # Escape regex special chars

    with pytest.raises(ValueError, match=escaped_msg):
        handle_if_empty(if_empty=invalid_value, message="This should fail.")  # type: ignore



from pathlib import Path

from viadot.orchestration.prefect.flows import duckdb_to_parquet
from viadot.sources import DuckDB


DUCKDB_CREDS = {"database": "test.duckdb", "read_only": False}
PATH = "test_parquet.parquet"


def test_duckdb_to_parquet():
    assert not Path(PATH).exists()

    duckdb = DuckDB(credentials=DUCKDB_CREDS)
    duckdb.run_query(
        query="""
--CREATE SCHEMA sandbox;
CREATE or replace TABLE sandbox.numbers AS
SELECT 42 AS i, 84 AS j;
"""
    )
    duckdb_to_parquet(
        query="""SELECT * FROM sandbox.numbers""",
        path=PATH,
        duckdb_credentials=DUCKDB_CREDS,
        if_exists="replace",
    )

    assert Path(PATH).exists()
    Path(PATH).unlink()



"""A module for interacting with MinIO."""

from collections.abc import Generator
from pathlib import Path
from typing import Literal

import pandas as pd
import pyarrow as pa
import pyarrow.parquet as pq


try:
    from minio import Minio
    from minio.error import S3Error
    import s3fs
except ModuleNotFoundError as e:
    msg = "Missing required modules to use MinIO source."
    raise ImportError(msg) from e

from pydantic import BaseModel
import urllib3
from urllib3.exceptions import NewConnectionError

from viadot.config import get_source_credentials
from viadot.sources.base import Source


class MinIOCredentials(BaseModel):
    endpoint: str = "localhost:9000"
    access_key: str
    secret_key: str
    bucket: str
    secure: bool = True  # Whether to use HTTPS.
    verify: bool = True  # Whether to verify TLS certificates.


class MinIO(Source):
    def __init__(
        self,
        credentials: MinIOCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """A class for interacting with MinIO.

        Interact with MinIO in a more Pythonic, user-friendly, and robust way than the
        official minio client.

        Args:
        credentials (MinIOCredentials): MinIO credentials.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = MinIOCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.endpoint = self.credentials.get("endpoint")
        self.access_key = self.credentials.get("access_key")
        self.secret_key = self.credentials.get("secret_key")
        self.bucket = self.credentials.get("bucket")
        self.secure = self.credentials.get("secure")
        self.verify = self.credentials.get("verify")

        self.http_scheme = "https" if self.secure is True else "http"

        self.client = Minio(
            self.endpoint,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure=self.secure,
            http_client=urllib3.PoolManager(
                timeout=urllib3.Timeout.DEFAULT_TIMEOUT,
                retries=urllib3.Retry(
                    connect=1,
                    read=3,
                    total=2,
                    backoff_factor=1,
                    status_forcelist=[500, 502, 503, 504],
                ),
                cert_reqs="CERT_REQUIRED" if self.verify else "CERT_NONE",
            ),
        )

        self.storage_options = {
            "key": self.access_key,
            "secret": self.secret_key,
            "client_kwargs": {
                "endpoint_url": f"{self.http_scheme}://" + self.endpoint,
                "use_ssl": self.secure,
                "verify": self.verify,
            },
        }

        self.fs = s3fs.S3FileSystem(
            key=self.storage_options["key"],
            secret=self.storage_options["secret"],
            client_kwargs=self.storage_options["client_kwargs"],
            anon=False,
        )

    def from_arrow(
        self,
        table: pa.Table,
        schema_name: str | None = None,
        table_name: str | None = None,
        path: str | Path | None = None,
        basename_template: str | None = None,
        partition_cols: list[str] | None = None,
        if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
    ) -> None:
        """Create a Parquet dataset on MinIO from a PyArrow Table.

        Uses multi-part upload to upload the table in chunks, speeding up the
        process by using multithreading and avoiding upload size limits.

        Either both `schema_name` and `table_name` or only `path` must be provided.

        `path` allows specifying an arbitrary path, while `schema_name` and `table_name`
        provide a shortcut for creating a data lakehouse-like structure of
        `s3://<bucket>/<schema_name>/<table_name>/<table_name>.parquet`.

        For more information on partitioning, see
        https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html#pyarrow-parquet-write-to-dataset

        Args:
            table (pa.Table): The table to upload.
            schema_name (str, optional): The name of the schema directory. Defaults to
                None.
            table_name (str, optional): The name of the table directory. Defaults to
                None.
            path (str | Path, optional): The path to the destination file. Defaults to
                None.
            basename_template (str, optional): A template string used to generate
                basenames of written data files. The token '{i}' will be replaced with
                an automatically incremented integer. Defaults to None.
            partition_cols (list[str], optional): The columns to partition by. Defaults
                to None.
            if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
                optional): What to do if the dataset already exists.
        """
        fqn_or_path = (schema_name and table_name) or (
            path and not (schema_name or table_name)
        )
        if not fqn_or_path:
            msg = "Either both `schema_name` and `table_name` or only `path` must be provided."
            raise ValueError(msg)

        # We need to create the dirs here as PyArrow also tries to create the bucket,
        # which shouldn't be allowed for whomever is executing this code.
        self.logger.debug(f"Creating directory for table {table_name}...")
        path = path or f"{schema_name}/{table_name}"
        self.fs.makedirs(path, exist_ok=True)
        self.logger.debug("Directory has been created successfully.")

        # Write the data.
        pq.write_to_dataset(
            table,
            root_path=path,
            partition_cols=partition_cols,
            existing_data_behavior=if_exists,
            basename_template=basename_template,
            filesystem=self.fs,
            max_rows_per_file=1024 * 1024,
            create_dir=False,  # Required as Arrow attempts to create the bucket, too.
        )

    def from_df(
        self,
        df: pd.DataFrame,
        schema_name: str | None = None,
        table_name: str | None = None,
        path: str | Path | None = None,
        basename_template: str | None = None,
        partition_cols: list[str] | None = None,
        if_exists: Literal["error", "delete_matching", "overwrite_or_ignore"] = "error",
    ) -> None:
        """Create a Parquet dataset on MinIO from a PyArrow Table.

        Uses multi-part upload to upload the table in chunks, speeding up the
        process by using multithreading and avoiding upload size limits.

        Either both `schema_name` and `table_name` or only `path` must be provided.

        `path` allows specifying an arbitrary path, while `schema_name` and `table_name`
        provide a shortcut for creating a data lakehouse-like structure of
        `s3://<bucket>/<schema_name>/<table_name>/<table_name>.parquet`.

        For more information on partitioning, see
        https://arrow.apache.org/docs/python/generated/pyarrow.parquet.write_to_dataset.html#pyarrow-parquet-write-to-dataset

        Args:
            df (pd.DataFrame): The DataFrame to upload.
            schema_name (str, optional): The name of the schema directory. Defaults to
                None.
            table_name (str, optional): The name of the table directory. Defaults to
                None.
            path (str | Path, optional): The path to the destination file. Defaults to
                None.
            basename_template (str, optional): A template string used to generate
                basenames of written data files. The token '{i}' will be replaced with
                an automatically incremented integer. Defaults to None.
            partition_cols (list[str], optional): The columns to partition by. Defaults
                to None.
            if_exists (Literal["error", "delete_matching", "overwrite_or_ignore"],
                optional): What to do if the dataset already exists.
        """
        table = pa.Table.from_pandas(df)

        return self.from_arrow(
            table=table,
            schema_name=schema_name,
            table_name=table_name,
            path=path,
            basename_template=basename_template,
            partition_cols=partition_cols,
            if_exists=if_exists,
        )

    def ls(self, path: str) -> Generator[str, None, None]:
        """List files and directories under `path`.

        List operation can be slow if there are a lot of objects, hence using a
        generator.

        Args:
            path (str): The path which contents should be listed.

        Yields:
            Generator[str, None, None]: Contents (files and directories) of `path`.
        """
        for obj in self.client.list_objects(self.bucket, path):
            yield obj.object_name

    def rm(self, path: str, recursive: bool = False) -> None:
        """Remove a file or directory from MinIO.

        Args:
            path (str): The path to the file to remove.
        """
        if recursive:
            bucket = path.split("/")[2]
            prefix = "/".join(path.split("/")[3:])
            objects = self.client.list_objects(bucket, prefix=prefix, recursive=True)
            for obj in objects:
                self.client.remove_object(bucket, obj.object_name)
            return

        self.client.remove_object(self.bucket, path)

    def _check_if_file_exists(self, path: str) -> bool:
        try:
            self.client.stat_object(self.bucket, path)
        except S3Error as e:
            if "Object does not exist" in e.message:
                return False
            raise
        else:
            return True

    def check_connection(self) -> None:
        """Verify connectivity to the MinIO endpoint."""
        try:
            self.client.bucket_exists(self.bucket)
        except NewConnectionError as e:
            msg = f"Connection to MinIO endpoint '{self.endpoint}' failed with error: \n{e}"
            msg += "Please check your credentials and try again."

            raise ValueError(msg) from e
        except Exception as e:
            msg = f"Connection to MinIO endpoint '{self.endpoint}' failed with error: \n{e}"
            raise ValueError(msg) from e
        self.logger.info("Connection successful!")



"""Flow for downloading data from TM1 to a Parquet file."""

from typing import Literal

from prefect import flow

from viadot.orchestration.prefect.tasks.task_utils import df_to_parquet
from viadot.orchestration.prefect.tasks.tm1 import tm1_to_df


@flow(
    name="extract--tm1--parquet",
    description="Extract data from TM1 and load it into Parquet file",
    retries=1,
    retry_delay_seconds=60,
)
def tm1_to_parquet(  # noqa: PLR0913
    path: str | None = None,
    mdx_query: str | None = None,
    cube: str | None = None,
    view: str | None = None,
    limit: int | None = None,
    private: bool = False,
    credentials_secret: str | None = None,
    config_key: str | None = None,
    if_empty: str = "skip",
    if_exists: Literal["append", "replace", "skip"] = "replace",
    verify: bool = True,
) -> None:
    """Download data from TM1 to a Parquet file.

    Args:
        mdx_query (str, optional): MDX select query needed to download the data.
            Defaults to None.
        cube (str, optional): Cube name from which data will be downloaded.
            Defaults to None.
        view (str, optional): View name from which data will be downloaded.
            Defaults to None.
        limit (str, optional): How many rows should be extracted.
            If None all the available rows will be downloaded. Defaults to None.
        private (bool, optional): Whether or not data download should be private.
            Defaults to False.
        credentials_secret (dict[str, Any], optional): The name of the secret that
            stores TM1 credentials.
            More info on: https://docs.prefect.io/concepts/blocks/. Defaults to None.
        config_key (str, optional): The key in the viadot config holding relevant
            credentials. Defaults to "TM1".
        if_empty (str, optional): What to do if output DataFrame is empty.
            Defaults to "skip".
        if_exists (Literal["append", "replace", "skip"], optional):
            What to do if the table exists. Defaults to "replace".
        verify (bool, optional): Whether or not verify SSL certificate.
            Defaults to False.
    """
    df = tm1_to_df(
        mdx_query=mdx_query,
        cube=cube,
        view=view,
        limit=limit,
        private=private,
        credentials_secret=credentials_secret,
        config_key=config_key,
        verify=verify,
        if_empty=if_empty,
    )
    return df_to_parquet(
        df=df,
        path=path,
        if_exists=if_exists,
    )



"""A module for interacting with Trino as a database."""

from collections.abc import Generator
from contextlib import contextmanager
import re
from typing import Literal
import warnings

import pandas as pd
import pyarrow as pa
from pydantic import BaseModel, Field
from sqlalchemy import create_engine, text
from sqlalchemy.engine import Connection, CursorResult
from sqlalchemy.exc import SADeprecationWarning
from trino.auth import BasicAuthentication

from viadot.config import get_source_credentials
from viadot.sources.base import Source
from viadot.utils import get_fqn


# Triggered by trino lib
warnings.filterwarnings("ignore", category=SADeprecationWarning)


class TrinoCredentials(BaseModel):
    """Trino credentials."""

    http_scheme: str = "https"
    host: str = "localhost"
    port: int = 443
    user: str
    password: str | None = None
    catalog: str
    schema_name: str | None = Field(None, alias="schema")
    verify: bool = True


class Trino(Source):
    def __init__(
        self,
        credentials: TrinoCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """A class for interacting with Trino as a database.

        Currently supports only generic and Iceberg operations.

        Args:
            credentials (TrinoCredentials): Trino credentials.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials.
        """
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = TrinoCredentials(**raw_creds).dict(
            by_alias=True
        )  # validate the credentials

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.http_scheme = self.credentials.get("http_scheme")
        self.host = self.credentials.get("host")
        self.port = self.credentials.get("port")
        self.username = self.credentials.get("user")
        self.password = self.credentials.get("password")
        self.catalog = self.credentials.get("catalog")
        self.schema = self.credentials.get("schema")
        self.verify = self.credentials.get("verify")

        self.connection_string = (
            f"trino://{self.username}@{self.host}:{self.port}/{self.catalog}"
        )
        self.connect_args = {
            "verify": self.verify,
            "auth": BasicAuthentication(self.username, self.password),
            "http_scheme": self.http_scheme,
        }
        self.engine = create_engine(
            self.connection_string, connect_args=self.connect_args, future=True
        )

    @contextmanager
    def get_connection(self) -> Generator[Connection, None, None]:
        """Provide a transactional scope around a series of operations.

        Examples:
        >>> trino = Trino()
        >>> with trino.get_connection() as connection:
        >>>    trino.run(query1, connection=connection)
        >>>    trino.run(query2, connection=connection)
        """
        connection = self.engine.connect()
        try:
            yield connection
            connection.commit()
        except:
            connection.rollback()
            raise
        finally:
            connection.close()

    def get_tables(self, schema_name: str) -> list[str]:
        """List all tables in a schema.

        Args:
            schema_name (str): _description_

        Returns:
            list[str]: _description_
        """
        query = f"SHOW TABLES FROM {schema_name}"
        with self.get_connection() as connection:
            return list(self.run(query, connection=connection))

    def drop_table(self, table_name: str, schema_name: str | None = None) -> None:
        """Drop a table.

        Args:
            table_name (str): _description_
            schema_name (str | None, optional): _description_. Defaults to None.
        """
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        query = f"DROP TABLE IF EXISTS {fqn}"

        self.logger.info(f"Dropping table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Table '{fqn}' has been successfully dropped.")

    def delete_table(self, table_name: str, schema_name: str | None = None) -> None:
        """Delete all data from a table.

        Args:
            table_name (str): _description_
            schema_name (str | None, optional): _description_. Defaults to None.
        """
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        query = f"DELETE FROM {fqn}"  # noqa: S608
        self.logger.info(f"Removing all data from table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Data from table '{fqn}' has been successfully removed.")

    def _check_if_table_exists(self, table_name: str, schema_name: str) -> None:
        query = f"""
SELECT *
FROM INFORMATION_SCHEMA.TABLES
WHERE TABLE_SCHEMA = '{schema_name}'
AND TABLE_NAME = '{table_name}'"""
        with self.get_connection() as connection:
            results = list(self.run(query, connection=connection))
        return len(results) > 0

    def get_schemas(self) -> list[str]:
        """List all schemas in the database.

        Returns:
            list[str]: _description_
        """
        query = "SHOW SCHEMAS"
        with self.get_connection() as connection:
            return list(self.run(query, connection=connection))

    def _check_if_schema_exists(self, schema_name: str) -> None:
        query = f"SELECT * FROM INFORMATION_SCHEMA.SCHEMATA WHERE SCHEMA_NAME = '{schema_name}'"  # noqa: S608
        with self.get_connection() as connection:
            results = list(self.run(query, connection=connection))
        return bool(results)

    def drop_schema(self, schema_name: str, cascade: bool = False) -> None:
        """Drop a schema.

        Args:
            schema_name (str): _description_
            cascade (bool, optional): _description_. Defaults to False.
        """
        if not self._check_if_schema_exists(schema_name):
            return

        if cascade:
            tables = self.get_tables(schema_name)
            [self.drop_table(schema_name=schema_name, table_name=t) for t in tables]

        self.logger.info(f"Dropping schema {schema_name}...")
        with self.get_connection() as connection:
            self.run(f"DROP SCHEMA {schema_name}", connection=connection)
        self.logger.info(f"Schema {schema_name} has been successfully dropped.")

    def create_iceberg_schema(
        self,
        schema_name: str,
        location: str,
        if_exists: Literal["fail", "skip"] = "fail",
    ) -> None:
        """Create an Iceberg schema.

        Args:
            schema_name (str): _description_
            location (str): _description_
            if_exists (Literal[&quot;fail&quot;, &quot;skip&quot;], optional): What
                to do if the schema already exists. Defaults to "fail".

        Raises:
            ValueError: _description_
        """
        exists = self._check_if_schema_exists(schema_name)

        if exists:
            if if_exists == "fail":
                msg = f"Schema '{schema_name}' already exists."
                raise ValueError(msg)

            self.logger.info(f"Schema '{schema_name}' already exists. Skipping...")
            return

        query = f"""
CREATE SCHEMA {schema_name}
WITH (location = '{location}')
        """
        self.logger.info(f"Creating schema '{schema_name}'...")
        with self.get_connection() as connection:
            self.run(query, connection=connection)
        self.logger.info(f"Schema '{schema_name}' has been successfully created.")

    def create_iceberg_table_from_arrow(
        self,
        table: pa.Table,
        table_name: str,
        schema_name: str | None = None,
        location: str | None = None,
        file_format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] | None = None,
        sort_by: list[str] | None = None,
    ) -> None:
        """Create an Iceberg table from a pyarrow Table.

        Args:
            table (pa.Table): _description_
            table_name (str): _description_
            schema_name (str | None, optional): _description_. Defaults to None.
            location (str | None, optional): _description_. Defaults to None.
            file_format (Literal[&quot;PARQUET&quot;, &quot;ORC&quot;], optional): The
                file format to use. Defaults to "PARQUET".
            partition_cols (list[str] | None, optional): The partition columns to use.
                Defaults to None.
            sort_by (list[str] | None, optional): _description_. Defaults to None.
        """
        columns = table.schema.names
        types = [self.pyarrow_to_trino_type(str(typ)) for typ in table.schema.types]
        create_table_query = self._create_table_query(
            schema_name=schema_name,
            table_name=table_name,
            columns=columns,
            types=types,
            location=location,
            file_format=file_format,
            partition_cols=partition_cols,
            sort_by_cols=sort_by,
        )

        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        self.logger.info(f"Creating table '{fqn}'...")
        with self.get_connection() as connection:
            self.run(create_table_query, connection=connection)
        self.logger.info(f"Table '{fqn}' has been successfully created.")

    def create_iceberg_table_from_pandas(
        self,
        df: pd.DataFrame,
        table_name: str,
        schema_name: str | None = None,
        location: str | None = None,
        file_format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] | None = None,
        sort_by: list[str] | None = None,
    ) -> None:
        """Create an Iceberg table from a pandas DataFrame.

        Args:
            df (pd.DataFrame): _description_
            table_name (str): _description_
            schema_name (str | None, optional): _description_. Defaults to None.
            location (str | None, optional): _description_. Defaults to None.
            file_format (Literal[&quot;PARQUET&quot;, &quot;ORC&quot;], optional): The
                file format to use. Defaults to "PARQUET".
            partition_cols (list[str] | None, optional): The partition columns to use.
                Defaults to None.
            sort_by (list[str] | None, optional): _description_. Defaults to None.
        """
        pa_table = pa.Table.from_pandas(df)
        self.create_iceberg_table_from_arrow(
            table=pa_table,
            schema_name=schema_name,
            table_name=table_name,
            location=location,
            file_format=file_format,
            partition_cols=partition_cols,
            sort_by=sort_by,
        )

    def _create_table_query(
        self,
        table_name: str,
        columns: list[str],
        types: list[str],
        schema_name: str | None = None,
        location: str | None = None,
        file_format: Literal["PARQUET", "ORC"] = "PARQUET",
        partition_cols: list[str] | None = None,
        sort_by_cols: list[str] | None = None,
    ) -> str:
        cols_and_dtypes = ",\n\t".join(
            col + " " + dtype for col, dtype in zip(columns, types, strict=False)
        )
        fqn = get_fqn(schema_name=schema_name, table_name=table_name)
        with_clause = f"format = '{file_format}'"

        if partition_cols:
            with_clause += ",\n\tpartitioning = ARRAY" + str(partition_cols)

        if sort_by_cols:
            with_clause += ",\n\tsorted_by = ARRAY" + str(sort_by_cols)

        if location:
            with_clause += f",\n\tlocation = '{location}'"

        return f"""
CREATE TABLE IF NOT EXISTS {fqn} (
    {cols_and_dtypes}
)
WITH (
    {with_clause}
)"""

    def run(
        self, sql: str, connection: Connection
    ) -> Generator[tuple, None, None] | None:
        """Run a SQL query.

        Args:
            sql (str): _description_
            connection (Connection): _description_

        Yields:
            Generator[tuple, None, None] | None: _description_
        """

        def row_generator(result: CursorResult):
            # Fetch rows in chunks of size `yield_per`.
            # This has to be inside a function due to how Python generators work.
            for partition in result.partitions():
                yield from partition

        self.logger.debug("Executing SQL:\n" + sql)

        try:
            # Execute with server-side cursor of size 5000.
            result = connection.execution_options(yield_per=5000).execute(text(sql))
        except Exception as e:
            msg = f"Failed executing SQL:\n{sql}"
            raise ValueError(msg) from e

        query_keywords = ["SELECT", "SHOW", "PRAGMA", "WITH"]
        is_query = any(sql.strip().upper().startswith(word) for word in query_keywords)

        return row_generator(result) if is_query else None

    @staticmethod
    def pyarrow_to_trino_type(pyarrow_type: str) -> str:
        """Convert a pyarrow data type to a Trino type.

        Args:
            pyarrow_type (str): The Pyarrow type to convert.

        Returns:
            str: The Trino type.
        """
        mapping = {
            "string": "VARCHAR",
            "large_string": "VARCHAR",
            "int8": "TINYINT",
            "int16": "SMALLINT",
            "int32": "INTEGER",
            "int64": "BIGINT",
            "float": "REAL",
            "double": "DOUBLE",
            "bool": "BOOLEAN",
            "date32[day]": "DATE",
            "timestamp[ns]": "TIMESTAMP(6)",
            "decimal": "DECIMAL",
            "decimal128": "DECIMAL",
            "decimal256": "DECIMAL",
        }
        precision, scale = None, None
        decimal_match = re.match(r"(\w+)\((\d+), (\d+)\)", pyarrow_type)
        if decimal_match:
            pyarrow_type = decimal_match.group(1)
            precision = int(decimal_match.group(2))
            scale = int(decimal_match.group(3))

        mapped_type = mapping.get(pyarrow_type) or "VARCHAR"

        if precision and scale:
            mapped_type += f"({precision}, {scale})"

        return mapped_type

    def _check_connection(self) -> None:
        try:
            with self.get_connection() as connection:
                self.run("select 1", connection=connection)
        except Exception as e:
            msg = f"Failed to connect to Trino server at {self.host}"
            raise ValueError(msg) from e



import os

from viadot.orchestration.prefect.flows import sap_to_redshift_spectrum
from viadot.sources import RedshiftSpectrum


S3_BUCKET = os.environ.get("S3_BUCKET")
TEST_SCHEMA = "raw_test"
TEST_TABLE = "test_sap_to_redshift_spectrum"


def test_sap_to_redshift_spectrum():
    sap_to_redshift_spectrum(
        to_path=f"s3://{S3_BUCKET}/nesso/{TEST_SCHEMA}/{TEST_TABLE}",
        schema_name=TEST_SCHEMA,
        table=TEST_TABLE,
        query="SELECT MATKL, MTART, ERSDA FROM MARA WHERE ERSDA = '20221230'",
        func="BBP_RFC_READ_TABLE",
        sap_config_key="sap_test",
    )

    rs = RedshiftSpectrum()

    received_df = rs.to_df(database=TEST_SCHEMA, table=TEST_TABLE)

    n_cols = 3
    assert not received_df.empty
    assert len(received_df.columns) == n_cols

    rs.rm(database=TEST_SCHEMA, table=TEST_TABLE)



from viadot.orchestration.prefect.flows import genesys_to_adls


def test_genesys_to_adls():
    state = genesys_to_adls(
        azure_key_vault_secret="genesys-access-1",  # noqa: S106
        verbose=True,
        endpoint="routing_queues_members",
        post_data_list=[""],
        queues_ids=[
            "25e29c3e-ba33-4556-a78b-2abc40ec9699",
            "f4ef329a-d903-41f4-ad4d-876a753adf3c",
        ],
        drop_duplicates=True,
        adls_azure_key_vault_secret="app-azure-cr-datalakegen2",  # noqa: S106
        adls_path="raw/dyvenia_sandbox/genesys/genesys_agents.csv",
        adls_path_overwrite=True,
    )
    assert state.is_successful()



"""Download data from Eurostat and upload it to Azure Data Lake Storage."""

from prefect import flow

from viadot.orchestration.prefect.tasks import df_to_adls, eurostat_to_df


@flow(
    name="extract--eurostat--adls",
    description="""Flow for downloading data from the Eurostat platform via
    HTTPS REST API (no credentials required) to a CSV or Parquet file.
    Then upload it to Azure Data Lake.""",
    retries=1,
    retry_delay_seconds=60,
)
def eurostat_to_adls(
    dataset_code: str,
    adls_path: str,
    params: dict[str, str] | None = None,
    columns: list[str] | None = None,
    tests: dict | None = None,
    adls_credentials_secret: str | None = None,
    overwrite_adls: bool = False,
    adls_config_key: str | None = None,
) -> None:
    """Flow for downloading data from Eurostat to Azure Data Lake.

    This module provides a prefect flow function to use the Eurostat connector:
    - Call the prefect task wrapper to retrieve a DataFrame from the connector.
    - Upload the retrieved data to Azure Data Lake Storage.

    Args:
        dataset_code (str):
            The code of the Eurostat dataset to be uploaded.
        adls_path (str | None, optional): The destination folder or path
            in Azure Data Lake.
        params (dict[str, str] | None, optional):
            A dictionary with optional URL parameters. Each key is a parameter ID,
            and the value is a specific parameter code, e.g.,
            `params = {'unit': 'EUR'}` where "unit" is the parameter, and "EUR"
            is the code. You can pass one code per parameter. Defaults to None.
        columns (list[str] | None, optional):
            A list of columns to filter from the DataFrame after downloading.
            This acts as a filter to retrieve only the needed columns. Defaults to None.
        tests (dict | None, optional):
            A dictionary containing test cases for the data, such as:
            - `column_size`: dict{column: size}
            - `column_unique_values`: list[columns]
            - `column_list_to_match`: list[columns]
            - `dataset_row_count`: dict{'min': number, 'max': number}
            - `column_match_regex`: dict{column: 'regex'}
            - `column_sum`: dict{column: {'min': number, 'max': number}}.
            Defaults to None.

        adls_credentials_secret (str | None, optional):
            The Azure Key Vault secret containing Service Principal credentials
            (TENANT_ID, CLIENT_ID, CLIENT_SECRET) and ACCOUNT_NAME for Azure Data
            Lake access. Defaults to None.
        overwrite_adls (bool, optional):
            Whether to overwrite files in the lake if they exist. Defaults to False.
        adls_config_key (str | None, optional):
            The key in the viadot config that holds the credentials for Azure
            Data Lake. Defaults to None.

    Returns:
        None
    """
    df = eurostat_to_df(
        dataset_code=dataset_code,
        params=params,
        columns=columns,
        tests=tests,
    )
    df_to_adls(
        df=df,
        path=adls_path,
        credentials_secret=adls_credentials_secret,
        config_key=adls_config_key,
        overwrite=overwrite_adls,
    )



"""Source for connecting to Epicor Prelude API."""

from typing import Any

import defusedxml.ElementTree as ET  # noqa: N817
import pandas as pd
from pydantic import BaseModel
import requests

from viadot.config import get_source_credentials
from viadot.exceptions import DataRangeError, ValidationError
from viadot.sources.base import Source
from viadot.utils import handle_api_response


"""The official documentation does not specify the list of required
fields so they were set as optional in BaseModel classes.

Each Epicor Prelude view requires different XML parser.
"""


class TrackingNumbers(BaseModel):
    TrackingNumber: str | None


class ShipToAddress(BaseModel):
    ShipToNumber: str | None
    Attention: str | None
    AddressLine1: str | None
    AddressLine2: str | None
    AddressLine3: str | None
    City: str | None
    State: str | None
    Zip: str | None
    Country: str | None
    EmailAddress: str | None
    PhoneNumber: str | None
    FaxNumber: str | None


class InvoiceTotals(BaseModel):
    Merchandise: str | None
    InboundFreight: str | None
    OutboundFreight: str | None
    Handling: str | None
    Delivery: str | None
    Pickup: str | None
    Restocking: str | None
    MinimumCharge: str | None
    DiscountAllowance: str | None
    SalesTax: str | None
    TotalInvoice: str | None


class HeaderInformation(BaseModel):
    CompanyNumber: str | None
    OrderNumber: str | None
    InvoiceNumber: str | None
    CustomerNumber: str | None
    CustomerDescription: str | None
    CustomerPurchaseOrderNumber: str | None
    Contact: str | None
    SellingWarehouse: str | None
    ShippingWarehouse: str | None
    ShippingMethod: str | None
    PaymentTerms: str | None
    PaymentTermsDescription: str | None
    FreightTerms: str | None
    FreightTermsDescription: str | None
    SalesRepOne: str | None
    SalesRepOneDescription: str | None
    EntryDate: str | None
    OrderDate: str | None
    RequiredDate: str | None
    ShippedDate: str | None
    InvoiceDate: str | None
    ShipToAddress: ShipToAddress | None
    TrackingNumbers: TrackingNumbers | None
    InvoiceTotals: InvoiceTotals | None


class LineItemDetail(BaseModel):
    ProductNumber: str | None
    ProductDescription: str | None
    ProductDescription1: str | None
    ProductDescription2: str | None
    CustomerProductNumber: str | None
    LineItemNumber: str | None
    QuantityOrdered: str | None
    QuantityShipped: str | None
    QuantityBackordered: str | None
    Price: str | None
    ExtendedCost: str | None
    ExtendedPrice: str | None
    QuantityShippedExtension: str | None
    LineItemShipWarehouse: str | None
    RequiredDate: str | None
    CopperWeight: str | None
    UnitOfMeasure: str | None
    Status: str | None
    GrossProfitExtension: str | None
    GrossProfitPercent: str | None
    UpdateDate: str | None
    DeleteReasonCode: str | None
    DeleteReasonDescription: str | None
    VendorProductNumber: str | None
    ProductGroup: str | None
    ProductGroupDescription: str | None
    ProductLine: str | None
    ProductLineDescription: str | None


class Customer(BaseModel):
    CompanyNumber: str | None
    CustomerNumber: str | None
    Description: str | None
    AddressOne: str | None
    AddressTwo: str | None
    AddressThree: str | None
    City: str | None
    State: str | None
    Zip: str | None
    Country: str | None
    Contact: str | None
    Phone: str | None
    EmailAddress: str | None
    ShipTos: ShipToAddress | None


class Order(BaseModel):
    HeaderInformation: HeaderInformation | None
    LineItemDetail: LineItemDetail | None


class BookingsInfo(BaseModel):
    HeaderInformation: HeaderInformation | None
    LineItemDetail: LineItemDetail | None


def parse_orders_xml(xml_data: str) -> pd.DataFrame:  # noqa: C901, PLR0912
    """Function to parse xml containing Epicor Orders Data.

    Args:
        xml_data (str, required): Response from Epicor API in form of xml

    Returns:
        pd.DataFrame: DataFrame containing parsed orders data.
    """
    dfs = []
    ship_dict = {}
    invoice_dict = {}
    header_params_dict = {}
    item_params_dict = {}

    root = ET.fromstring(xml_data.text)

    for order in root.findall("Order"):
        for header in order.findall("HeaderInformation"):
            for tracking_numbers in header.findall("TrackingNumbers"):
                numbers = ""
                for tracking_number in tracking_numbers.findall("TrackingNumber"):
                    numbers = numbers + "'" + tracking_number.text + "'"
                result_numbers = TrackingNumbers(TrackingNumber=numbers)

            for shipto in header.findall("ShipToAddress"):
                for ship_param in ShipToAddress.__dict__.get("__annotations__"):
                    try:
                        ship_value = shipto.find(ship_param).text
                    except AttributeError:
                        ship_value = None
                    ship_parameter = {ship_param: ship_value}
                    ship_dict.update(ship_parameter)
                ship_address = ShipToAddress(**ship_dict)

            for invoice in header.findall("InvoiceTotals"):
                for invoice_param in InvoiceTotals.__dict__.get("__annotations__"):
                    try:
                        invoice_value = invoice.find(invoice_param).text
                    except AttributeError:
                        invoice_value = None
                    invoice_parameter = {invoice_param: invoice_value}
                    invoice_dict.update(invoice_parameter)
                invoice_total = InvoiceTotals(**invoice_dict)

            for header_param in HeaderInformation.__dict__.get("__annotations__"):
                try:
                    header_value = header.find(header_param).text
                except AttributeError:
                    header_value = None
                if header_param == "TrackingNumbers":
                    header_parameter = {header_param: result_numbers}
                elif header_param == "ShipToAddress":
                    header_parameter = {header_param: ship_address}
                elif header_param == "InvoiceTotals":
                    header_parameter = {header_param: invoice_total}
                else:
                    header_parameter = {header_param: header_value}
                header_params_dict.update(header_parameter)
            header_info = HeaderInformation(**header_params_dict)
        for items in order.findall("LineItemDetails"):
            for item in items.findall("LineItemDetail"):
                for item_param in LineItemDetail.__dict__.get("__annotations__"):
                    try:
                        item_value = item.find(item_param).text
                    except AttributeError:
                        item_value = None
                    item_parameter = {item_param: item_value}
                    item_params_dict.update(item_parameter)
                line_item = LineItemDetail(**item_params_dict)
                row = Order(HeaderInformation=header_info, LineItemDetail=line_item)
                my_dict = row.dict()
                dfs.append(pd.json_normalize(my_dict, max_level=2))
    return pd.concat(dfs, ignore_index=True)


def parse_bookings_xml(xml_data: str) -> pd.DataFrame:  # noqa: C901, PLR0912
    """Function to parse xml containing Epicor Bookings Data.

    Args:
        xml_data (str, required): Response from Epicor API in form of xml
    Returns:
        pd.DataFrame: DataFrame containing parsed  data.
    """
    dfs = []
    ship_dict = {}
    header_params_dict = {}
    item_params_dict = {}
    root = ET.fromstring(xml_data.text)

    for booking in root.findall("BookingsInfo"):
        for header in booking.findall("HeaderInformation"):
            for shipto in header.findall("ShipToAddress"):
                for ship_param in ShipToAddress.__dict__.get("__annotations__"):
                    try:
                        ship_value = shipto.find(ship_param).text
                    except AttributeError:
                        ship_value = None
                    ship_parameter = {ship_param: ship_value}
                    ship_dict.update(ship_parameter)
                ship_address = ShipToAddress(**ship_dict)

            for header_param in HeaderInformation.__dict__.get("__annotations__"):
                try:
                    header_value = header.find(header_param).text
                except AttributeError:
                    header_value = None
                if header_param == "ShipToAddress":
                    header_parameter = {header_param: ship_address}
                else:
                    header_parameter = {header_param: header_value}
                header_params_dict.update(header_parameter)
            header_info = HeaderInformation(**header_params_dict)
        for items in booking.findall("LineItemDetails"):
            for item in items.findall("LineItemDetail"):
                for item_param in LineItemDetail.__dict__.get("__annotations__"):
                    try:
                        item_value = item.find(item_param).text
                    except AttributeError:
                        item_value = None
                    item_parameter = {item_param: item_value}
                    item_params_dict.update(item_parameter)
                line_item = LineItemDetail(**item_params_dict)
                row = BookingsInfo(
                    HeaderInformation=header_info, LineItemDetail=line_item
                )
                my_dict = row.dict()
                dfs.append(pd.json_normalize(my_dict, max_level=2))
    return pd.concat(dfs, ignore_index=True)


def parse_open_orders_xml(xml_data: str) -> pd.DataFrame:  # noqa: C901, PLR0912
    """Function to parse xml containing Epicor Open Orders Data.

    Args:
        xml_data (str, required): Response from Epicor API in form of xml
    Returns:
        pd.DataFrame: DataFrame containing parsed  data.
    """
    dfs = []
    ship_dict = {}
    header_params_dict = {}
    item_params_dict = {}

    root = ET.fromstring(xml_data.text)

    for order in root.findall("Order"):
        for header in order.findall("HeaderInformation"):
            for shipto in header.findall("ShipToAddress"):
                for ship_param in ShipToAddress.__dict__.get("__annotations__"):
                    try:
                        ship_value = shipto.find(ship_param).text
                    except AttributeError:
                        ship_value = None
                    ship_parameter = {ship_param: ship_value}
                    ship_dict.update(ship_parameter)
                ship_address = ShipToAddress(**ship_dict)

            for header_param in HeaderInformation.__dict__.get("__annotations__"):
                try:
                    header_value = header.find(header_param).text
                except AttributeError:
                    header_value = None
                if header_param == "ShipToAddress":
                    header_parameter = {header_param: ship_address}
                else:
                    header_parameter = {header_param: header_value}
                header_params_dict.update(header_parameter)
            header_info = HeaderInformation(**header_params_dict)
        for items in order.findall("LineItemDetails"):
            for item in items.findall("LineItemDetail"):
                for item_param in LineItemDetail.__dict__.get("__annotations__"):
                    try:
                        item_value = item.find(item_param).text
                    except AttributeError:
                        item_value = None
                    item_parameter = {item_param: item_value}
                    item_params_dict.update(item_parameter)
                line_item = LineItemDetail(**item_params_dict)
                row = Order(HeaderInformation=header_info, LineItemDetail=line_item)
                my_dict = row.dict()
                dfs.append(pd.json_normalize(my_dict, max_level=2))
    return pd.concat(dfs, ignore_index=True)


def parse_customer_xml(xml_data: str) -> pd.DataFrame:
    """Function to parse xml containing Epicor Customers Data.

    Args:
        xml_data (str, required): Response from Epicor API in form of xml
    Returns:
        pd.DataFrame: DataFrame containing parsed  data.
    """
    dfs = []
    ship_dict = {}
    customer_params_dict = {}

    root = ET.fromstring(xml_data.text)

    for customer in root.findall("Customer"):
        for ship in customer.findall("ShipTos"):
            for shipto in ship.findall("ShipToAddress"):
                for ship_param in ShipToAddress.__dict__.get("__annotations__"):
                    try:
                        ship_value = shipto.find(ship_param).text
                    except AttributeError:
                        ship_value = None
                    ship_parameter = {ship_param: ship_value}
                    ship_dict.update(ship_parameter)
                ship_address = ShipToAddress(**ship_dict)

                for cust_param in Customer.__dict__.get("__annotations__"):
                    try:
                        cust_value = customer.find(cust_param).text
                    except AttributeError:
                        cust_value = None
                    if cust_param == "ShipTos":
                        cust_parameter = {cust_param: ship_address}
                    else:
                        cust_parameter = {cust_param: cust_value}
                    customer_params_dict.update(cust_parameter)
                cust_info = Customer(**customer_params_dict)
                my_dict = cust_info.dict()
                dfs.append(pd.json_normalize(my_dict, max_level=2))
    return pd.concat(dfs, ignore_index=True)


class EpicorCredentials(BaseModel):
    host: str
    port: int = 443
    username: str
    password: str


class Epicor(Source):
    def __init__(
        self,
        base_url: str,
        credentials: dict[str, Any] | None = None,
        config_key: str | None = None,
        validate_date_filter: bool = True,
        start_date_field: str = "BegInvoiceDate",
        end_date_field: str = "EndInvoiceDate",
        *args,
        **kwargs,
    ):
        """Class to connect to Epicor API and pasere XML output into a pandas DataFrame.

        Args:
            base_url (str, required): Base url to Epicor.
            filters_xml (str, required): Filters in form of XML. The date filter
                 is required.
            credentials (dict[str, Any], optional): Credentials to connect with
            Epicor API containing host, port, username and password.Defaults to None.
            config_key (str, optional): Credential key to dictionary where details
                 are stored.
            validate_date_filter (bool, optional): Whether or not validate xml
                date filters. Defaults to True.
            start_date_field (str, optional) The name of filters field containing
                start date.Defaults to "BegInvoiceDate".
            end_date_field (str, optional) The name of filters field containing
                end date. Defaults to "EndInvoiceDate".
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(EpicorCredentials(**raw_creds))

        self.base_url = base_url
        self.validate_date_filter = validate_date_filter
        self.start_date_field = start_date_field
        self.end_date_field = end_date_field

        self.url = (
            "http://"
            + validated_creds["host"]
            + ":"
            + str(validated_creds["port"])
            + base_url
        )

        super().__init__(*args, credentials=validated_creds, **kwargs)

    def generate_token(self) -> str:
        """Function to generate API access token that is valid for 24 hours.

        Returns:
            str: Generated token.
        """
        url = (
            "http://"
            + self.credentials["host"]
            + ":"
            + str(self.credentials["port"])
            + "/api/security/token/"
        )

        headers = {
            "Content-Type": "application/json",
            "username": self.credentials["username"],
            "password": self.credentials["password"],
        }

        response = handle_api_response(url=url, headers=headers, method="POST")
        root = ET.fromstring(response.text)
        return root.find("AccessToken").text

    def validate_filter(self, filters_xml: str) -> None:
        "Function checking if user had specified date range filters."
        root = ET.fromstring(filters_xml)
        for child in root:
            for subchild in child:
                if (
                    subchild.tag in (self.start_date_field, self.end_date_field)
                ) and subchild.text is None:
                    msg = "Too much data. Please provide a date range filter."
                    raise DataRangeError(msg)

    def get_xml_response(self, filters_xml: str) -> requests.models.Response:
        "Function for getting response from Epicor API."
        if self.validate_date_filter is True:
            self.validate_filter(filters_xml)
        payload = filters_xml
        headers = {
            "Content-Type": "application/xml",
            "Authorization": "Bearer " + self.generate_token(),
        }
        return handle_api_response(
            url=self.url, headers=headers, data=payload, method="POST"
        )

    def to_df(self, filters_xml: str) -> pd.DataFrame:
        """Function for creating pandas DataFrame from Epicor API response.

        Returns:
            pd.DataFrame: Output DataFrame.
        """
        data = self.get_xml_response(filters_xml)
        if "ORDER.HISTORY.DETAIL.QUERY" in self.base_url:
            df = parse_orders_xml(data)
        elif "CUSTOMER.QUERY" in self.base_url:
            df = parse_customer_xml(data)
        elif "ORDER.DETAIL.PROD.QUERY" in self.base_url:
            df = parse_open_orders_xml(data)
        elif "BOOKINGS.DETAIL.QUERY" in self.base_url:
            df = parse_bookings_xml(data)
        else:
            msg = f"Parser for selected viev {self.base_url} is not avaiable"
            raise ValidationError(msg)

        return df



import contextlib
import os

import pandas as pd
import pytest

from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import S3
except ImportError:
    skip_test_on_missing_extra(source_name="S3", extra="aws")

from pathlib import Path


SOURCE_DATA = [
    {
        "Name": "Scott-Merritt",
        "FirstName": "Melody",
        "LastName": "Cook",
        "ContactEmail": "Melody.Cook@ScottMerritt.com",
        "MailingCity": "Elizabethfurt",
    },
    {
        "Name": "Mann-Warren",
        "FirstName": "Wayne",
        "LastName": "Morrison",
        "ContactEmail": "Wayne.Morrison@MannWarren.com",
        "MailingCity": "Kathrynmouth",
    },
]
TEST_DF = pd.DataFrame(SOURCE_DATA)

S3_BUCKET = os.environ.get("VIADOT_S3_BUCKET")
TEST_SCHEMA = "test_schema"
TEST_TABLE = "test_table"
TEST_TABLE_PATH = f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}"
TEST_TABLE_PATH_PARQUET = str(Path(TEST_TABLE_PATH) / f"{TEST_TABLE}.parquet")
TEST_TABLE_PATH_CSV = str(Path(TEST_TABLE_PATH) / f"{TEST_TABLE}.csv")


@pytest.fixture(scope="session")
def s3(s3_config_key):
    s3 = S3(config_key=s3_config_key)

    yield s3

    # Remove the s3 table folder.
    with contextlib.suppress(Exception):
        s3.rm(path=TEST_TABLE_PATH)


def test_from_df(s3):
    # Assumptions.
    exists = s3.exists(path=TEST_TABLE_PATH)
    assert exists is False

    # Test.
    s3.from_df(df=TEST_DF, path=TEST_TABLE_PATH)
    result = s3.exists(path=TEST_TABLE_PATH)
    assert result is True

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_from_df_max_rows(s3):
    rows_per_file = 1
    size_df = len(TEST_DF)
    amount_of_expected_files = int(size_df / rows_per_file)

    s3.from_df(
        df=TEST_DF,
        path=TEST_TABLE_PATH,
        max_rows_by_file=rows_per_file,
    )

    counter_files = 0

    for i in range(0, amount_of_expected_files, 1):
        if s3.exists(path=f"{TEST_TABLE_PATH}_{i}"):
            counter_files = counter_files + 1
            s3.rm(path=[f"{TEST_TABLE_PATH}_{i}"])

    assert (counter_files == amount_of_expected_files) is True

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_to_df(s3):
    # Assumptions.
    exists = s3.exists(path=TEST_TABLE_PATH_PARQUET)
    assert exists is False

    # Test.
    s3.from_df(df=TEST_DF, path=TEST_TABLE_PATH_PARQUET)
    result = s3.to_df(paths=[TEST_TABLE_PATH_PARQUET])
    assert len(result) == len(TEST_DF)

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_to_df_chunk_size(s3):
    # Assumptions.
    exists = s3.exists(path=TEST_TABLE_PATH_PARQUET)
    assert exists is False

    # Test.
    s3.from_df(df=TEST_DF, path=TEST_TABLE_PATH_PARQUET)

    dfs = s3.to_df(
        paths=[TEST_TABLE_PATH_PARQUET],
        chunk_size=1,
    )

    result = pd.concat(dfs, ignore_index=True)
    assert len(result) == len(TEST_DF)

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_upload(s3, TEST_CSV_FILE_PATH):
    # Assumptions.
    exists = s3.exists(path=TEST_TABLE_PATH_CSV)
    assert exists is False

    # Test.
    s3.upload(
        from_path=TEST_CSV_FILE_PATH,
        to_path=TEST_TABLE_PATH_CSV,
    )

    exists = s3.exists(path=TEST_TABLE_PATH_CSV)
    assert exists is True

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_download(s3, TEST_CSV_FILE_PATH):
    # Assumptions.
    assert not Path("test.csv").exists()

    s3.upload(
        from_path=TEST_CSV_FILE_PATH,
        to_path=TEST_TABLE_PATH_CSV,
    )
    exists = s3.exists(path=TEST_TABLE_PATH_CSV)
    assert exists is True

    # Test.
    s3.download(
        from_path=TEST_TABLE_PATH_CSV,
        to_path="test.csv",
    )

    assert Path("test.csv").exists()

    # Cleanup.
    Path("test.csv").unlink()
    s3.rm(path=TEST_TABLE_PATH)


def test_if_exists(s3, TEST_CSV_FILE_PATH):
    """Test that `S3.exists()` works for files."""
    # Assumptions.
    exists = s3.exists(path=TEST_TABLE_PATH_CSV)
    assert exists is False

    # Test.
    s3.upload(
        from_path=TEST_CSV_FILE_PATH,
        to_path=TEST_TABLE_PATH_CSV,
    )
    exists = s3.exists(path=TEST_TABLE_PATH_CSV)
    assert exists is True

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)


def test_if_exists_folder(s3, TEST_CSV_FILE_PATH):
    """Test that `S3.exists()` works for folders."""
    # Assumptions.
    exists = s3.exists(TEST_TABLE_PATH)
    assert exists is False

    # Test.
    s3.upload(
        from_path=TEST_CSV_FILE_PATH,
        to_path=TEST_TABLE_PATH_CSV,
    )
    exists = s3.exists(TEST_TABLE_PATH)
    assert exists is True

    # Cleanup.
    s3.rm(path=TEST_TABLE_PATH)



from datetime import datetime
from io import StringIO
import json
import unittest
from unittest.mock import MagicMock, patch

import pandas as pd
import pytest
from requests.models import Response

from viadot.exceptions import APIError
from viadot.sources import Hubspot
from viadot.sources.hubspot import HubspotCredentials


variables = {
    "credentials": {"token": "fake_token"},
    "filters": [
        {
            "filters": [
                {
                    "propertyName": "createdate",
                    "operator": "GTE",
                    "value": "2021-01-01",
                }
            ]
        }
    ],
}


class TestHubspotCredentials:
    """Test Hubspot Credentials Class."""

    def test_hubspot_credentials(self):
        """Test Hubspot credentials."""
        HubspotCredentials(token="test_token")  # noqa: S106


class TestHubspot(unittest.TestCase):
    """Test Hubspot Class."""

    @classmethod
    def setUpClass(cls):  # noqa: ANN206
        """Defined based Hubspot Class for the rest of test."""
        cls.hubspot_instance = Hubspot(credentials=variables["credentials"])

    @patch("viadot.sources.hubspot.get_source_credentials", return_value=None)
    def test_init_no_credentials(self, mock_get_source_credentials):
        """Test raise error without credentials."""
        with pytest.raises(TypeError):
            Hubspot()

        mock_get_source_credentials.assert_called_once()

    def test_date_to_unixtimestamp(self):
        """Test Hubspot `_date_to_unixtimestamp` function."""
        date_str = "2021-01-01"
        expected_timestamp = int(
            datetime.strptime(date_str, "%Y-%m-%d").timestamp() * 1000
        )
        result = self.hubspot_instance._date_to_unixtimestamp(date_str)
        assert result == expected_timestamp

    def test_get_api_url(self):
        """Test Hubspot `_get_api_url` function."""
        endpoint = "deals"
        filters = None
        properties = ["property1", "property2"]
        expected_url = (
            f"https://api.hubapi.com/crm/v3/objects/{endpoint}/"
            + "?limit=100&properties=property1,property2&"
        )
        result = self.hubspot_instance._get_api_url(
            endpoint=endpoint, filters=filters, properties=properties
        )
        assert result == expected_url

    def test_format_filters(self):
        """Test Hubspot `_format_filters` function."""
        filters = variables["filters"]
        formatted_filters = self.hubspot_instance._format_filters(filters)
        assert isinstance(formatted_filters, list)

    def test_get_api_body(self):
        """Test Hubspot `_get_api_body` function."""
        filters = variables["filters"]
        expected_body = json.dumps({"filterGroups": filters, "limit": 100})
        result = self.hubspot_instance._get_api_body(filters)

        assert result == expected_body

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_call_success(self, mock_handle_api_response):
        """Test Hubspot `_api_call` method."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        url = "https://api.hubapi.com/crm/v3/objects/deals/?limit=100&"
        result = self.hubspot_instance._api_call(url=url, method="GET")

        assert result == {"results": [{"id": "123"}]}
        mock_handle_api_response.assert_called_once()

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_call_error(self, mock_handle_api_response):
        """Test Hubspot `_api_call` method failure."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 500
        mock_response.content = b"Internal Server Error"
        mock_handle_api_response.return_value = mock_response

        url = "https://api.hubapi.com/crm/v3/objects/deals/?limit=100&"

        with pytest.raises(APIError):
            self.hubspot_instance._api_call(url=url, method="GET")

    def test_get_offset_from_response(self):
        """Test Hubspot `_get_offset_from_response` function."""
        response_with_paging = {"paging": {"next": {"after": "123"}}}
        response_with_offset = {"offset": "456"}

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response(
            response_with_paging
        )
        assert (offset_type, offset_value) == ("after", "123")

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response(
            response_with_offset
        )
        assert (offset_type, offset_value) == ("offset", "456")

        offset_type, offset_value = self.hubspot_instance._get_offset_from_response({})
        assert (offset_type, offset_value) == (None, None)

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_connection_with_filters(self, mock_handle_api_response):
        """Test Hubspot `api_connection` method, with filters."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        endpoint = "deals"
        filters = variables["filters"]
        properties = ["property1"]
        self.hubspot_instance.api_connection(
            endpoint=endpoint, filters=filters, properties=properties
        )

        assert self.hubspot_instance.full_dataset is not None
        assert len(self.hubspot_instance.full_dataset) > 0

    @patch("viadot.sources.hubspot.handle_api_response")
    def test_api_connection_without_filters(self, mock_handle_api_response):
        """Test Hubspot `api_connection` method, without filters."""
        mock_response = MagicMock(spec=Response)
        mock_response.status_code = 200
        mock_response.json.return_value = {"results": [{"id": "123"}]}
        mock_handle_api_response.return_value = mock_response

        endpoint = "deals"
        filters = None
        properties = ["property1"]
        self.hubspot_instance.api_connection(
            endpoint=endpoint, filters=filters, properties=properties
        )

        assert self.hubspot_instance.full_dataset is not None
        assert len(self.hubspot_instance.full_dataset) > 0

    @patch("viadot.sources.hubspot.super")
    def test_to_df(self, mock_super):
        """Test Hubspot `to_df` function."""
        mock_super().to_df = MagicMock()
        self.hubspot_instance.full_dataset = [{"id": "123"}]
        result_df = self.hubspot_instance.to_df()
        result_df.drop(
            columns=["_viadot_source", "_viadot_downloaded_at_utc"],
            inplace=True,
            axis=1,
        )

        expected_df = pd.DataFrame([{"id": "123"}])
        assert result_df.equals(expected_df)
        mock_super().to_df.assert_called_once()

    @patch("viadot.sources.hubspot.pd.read_json")
    @patch("viadot.sources.hubspot.super")
    def test_to_df_empty(self, mock_super, mock_read_json):
        """Test Hubspot `to_df` method, checking emptiness."""
        mock_super().to_df = MagicMock()
        mock_read_json.return_value = pd.DataFrame()
        self.hubspot_instance.full_dataset = StringIO("{}")

        with patch.object(
            self.hubspot_instance, "_handle_if_empty"
        ) as mock_handle_if_empty:
            result_df = self.hubspot_instance.to_df()
            result_df.drop(
                columns=["_viadot_source", "_viadot_downloaded_at_utc"],
                inplace=True,
                axis=1,
            )
            mock_handle_if_empty.assert_called_once_with(
                if_empty="warn", message="The response does not contain any data."
            )
            assert result_df.empty
            mock_super().to_df.assert_called_once()


if __name__ == "__main__":
    unittest.main()



import contextlib
import os

import pandas as pd
import pytest

from viadot.utils import skip_test_on_missing_extra


try:
    from viadot.sources import S3, RedshiftSpectrum
except ImportError:
    skip_test_on_missing_extra(source_name="RedshiftSpectrum", extra="aws")


TEST_DF = pd.DataFrame(
    [
        [0, "A"],
        [1, "B"],
        [2, "C"],
    ],
    columns=["col1", "col2"],
)
TEST_DF = TEST_DF.astype({"col1": "Int32", "col2": "string"})

S3_BUCKET = os.environ.get("VIADOT_S3_BUCKET")
TEST_SCHEMA = "test_schema"
TEST_TABLE = "test_table"


@pytest.fixture(scope="session")
def redshift(redshift_config_key):
    redshift = RedshiftSpectrum(config_key=redshift_config_key)

    yield redshift

    with contextlib.suppress(Exception):
        redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)

    with contextlib.suppress(AttributeError):
        redshift._con.close()


@pytest.fixture(scope="session")
def s3(s3_config_key):
    return S3(config_key=s3_config_key)


def test_create_schema(redshift):
    # Assumptions.
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is False

    # Test.
    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    # Cleanup.
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test__check_if_table_exists(redshift):
    # Assumptions.
    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    table_exists = redshift._check_if_table_exists(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    assert table_exists is False

    # Test.
    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    table_exists = redshift._check_if_table_exists(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    assert table_exists is True

    # Cleanup.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)
    redshift.drop_schema(TEST_SCHEMA)


def test_from_df(redshift, s3):
    # Assumptions.
    table_exists = redshift._check_if_table_exists(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    assert table_exists is False

    folder_exists = s3.exists(
        path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}"
    )
    assert folder_exists is False

    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    # Test.
    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    table_exists = redshift._check_if_table_exists(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )

    assert table_exists is True

    folder_exists = s3.exists(
        path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}"
    )
    assert folder_exists is True

    # Cleanup.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test_from_df_no_table_folder_in_to_path(redshift, s3):
    """Test that the table folder is created if it's not specified in `to_path`."""
    # Assumptions.
    table_exists = redshift._check_if_table_exists(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    assert table_exists is False

    folder_exists = s3.exists(
        path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}"
    )
    assert folder_exists is False

    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    # Test.
    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )

    table_exists = redshift._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)
    assert table_exists is True

    folder_exists = s3.exists(
        path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}"
    )
    assert folder_exists is True

    # Cleanup.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test_to_df(redshift):
    # Assumptions.
    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    # Test.
    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )

    df = redshift.to_df(
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    assert df.equals(TEST_DF)

    # Cleanup.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test_get_tables(redshift):
    # Assumptions.
    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    tables = redshift.get_tables(schema=TEST_SCHEMA)
    assert TEST_TABLE not in tables

    # Test.
    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    tables = redshift.get_tables(schema=TEST_SCHEMA)
    assert tables == [TEST_TABLE]

    # Cleanup.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test_get_tables_incorrect_schema(redshift):
    """Verify that `get_tables()` returns an empty list if the schema doesn't exist."""
    # Assumptions.
    schema_exists = redshift._check_if_schema_exists("test_schema2")
    assert schema_exists is False

    # Test.
    tables = redshift.get_tables(schema="test_schema2")
    assert tables == []


def test_drop_table(redshift):
    # Assumptions.
    redshift.create_schema(TEST_SCHEMA)
    schema_exists = redshift._check_if_schema_exists(schema=TEST_SCHEMA)
    assert schema_exists is True

    redshift.from_df(
        df=TEST_DF,
        to_path=f"s3://{S3_BUCKET}/viadot/{TEST_SCHEMA}/{TEST_TABLE}",
        schema=TEST_SCHEMA,
        table=TEST_TABLE,
    )
    table_exists = redshift._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)
    assert table_exists is True

    # Test.
    redshift.drop_table(schema=TEST_SCHEMA, table=TEST_TABLE, remove_files=True)

    table_exists = redshift._check_if_table_exists(schema=TEST_SCHEMA, table=TEST_TABLE)
    assert table_exists is False

    # Cleanup.
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)


def test__is_spectrum_schema(redshift):
    # Prerequisite.
    redshift.create_schema(TEST_SCHEMA)

    # Test.
    assert not redshift._is_spectrum_schema("public")
    assert redshift._is_spectrum_schema(TEST_SCHEMA)

    # Cleanup.
    redshift.drop_schema(TEST_SCHEMA, drop_glue_database=True)



"""Task for downloading data from Hubspot API to a pandas DataFrame."""

from typing import Any

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.exceptions import MissingSourceCredentialsError
from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import Hubspot


@task(retries=3, log_prints=True, retry_delay_seconds=10, timeout_seconds=60 * 60)
def hubspot_to_df(
    endpoint: str,
    config_key: str | None = None,
    azure_key_vault_secret: str | None = None,
    filters: list[dict[str, Any]] | None = None,
    properties: list[Any] | None = None,
    nrows: int = 1000,
) -> pd.DataFrame:
    """Task to download data from Hubspot API to a pandas DataFrame.

    Args:
        endpoint (str): API endpoint for an individual request.
        config_key (Optional[str], optional): The key in the viadot config holding
            relevant credentials. Defaults to None.
        azure_key_vault_secret (Optional[str], optional): The name of the Azure Key
            Vault secret where credentials are stored. Defaults to None.
        filters (Optional[List[Dict[str, Any]]], optional): Filters defined for the API
            body in specific order. Defaults to None.
        properties (Optional[List[Any]], optional): List of user-defined columns to be
            pulled from the API. Defaults to None.
        nrows (int, optional): Max number of rows to pull during execution.
            Defaults to 1000.

    Examples:
        data_frame = hubspot_to_df(
            config_key=config_key,
            azure_key_vault_secret=azure_key_vault_secret,
            endpoint=endpoint,
            filters=filters,
            properties=properties,
            nrows=nrows,
        )

    Raises:
        MissingSourceCredentialsError: If no credentials have been provided.

    Returns:
        pd.DataFrame: The response data as a pandas DataFrame.
    """
    if not (azure_key_vault_secret or config_key):
        raise MissingSourceCredentialsError

    if not config_key:
        credentials = get_credentials(azure_key_vault_secret)

    hubspot = Hubspot(
        credentials=credentials,
        config_key=config_key,
    )
    hubspot.api_connection(
        endpoint=endpoint,
        filters=filters,
        properties=properties,
        nrows=nrows,
    )

    return hubspot.to_df()



"""Task for downloading data from Azure SQL."""

from typing import Any, Literal

import pandas as pd
from prefect import task

from viadot.orchestration.prefect.utils import get_credentials
from viadot.sources import AzureSQL
from viadot.utils import validate


@task(retries=3, retry_delay_seconds=10, timeout_seconds=60 * 60)
def azure_sql_to_df(
    query: str | None = None,
    credentials_secret: str | None = None,
    validate_df_dict: dict[str, Any] | None = None,
    convert_bytes: bool = False,
    remove_special_characters: bool | None = None,
    columns_to_clean: list[str] | None = None,
    if_empty: Literal["warn", "skip", "fail"] = "warn",
) -> pd.DataFrame:
    r"""Task to download data from Azure SQL.

    Args:
        query (str): Query to perform on a database. Defaults to None.
        credentials_secret (str, optional): The name of the Azure Key Vault
            secret containing a dictionary with database credentials.
            Defaults to None.
        validate_df_dict (Dict[str], optional): A dictionary with optional list of
            tests to verify the output dataframe. If defined, triggers the `validate_df`
            task from task_utils. Defaults to None.
        convert_bytes (bool). A boolean value to trigger method df_converts_bytes_to_int
            It is used to convert bytes data type into int, as pulling data with bytes
            can lead to malformed data in data frame.
            Defaults to False.
        remove_special_characters (str, optional): Call a function that remove
            special characters like escape symbols. Defaults to None.
        columns_to_clean (List(str), optional): Select columns to clean, used with
            remove_special_characters. If None whole data frame will be processed.
            Defaults to None.
        if_empty (Literal["warn", "skip", "fail"], optional): What to do if the
            query returns no data. Defaults to None.

    Raises:
        ValueError: Raising ValueError if credentials_secret is not provided

    Returns:
        pd.DataFrame: The response data as a pandas DataFrame.
    """
    if not credentials_secret:
        msg = "`credentials_secret` has to be specified and not empty."
        raise ValueError(msg)

    credentials = get_credentials(credentials_secret)

    azure_sql = AzureSQL(credentials=credentials)

    df = azure_sql.to_df(
        query=query,
        if_empty=if_empty,
        convert_bytes=convert_bytes,
        remove_special_characters=remove_special_characters,
        columns_to_clean=columns_to_clean,
    )

    if validate_df_dict is not None:
        validate(df=df, tests=validate_df_dict)

    return df



"""SMB (file-sharing protocol) connector."""

from pathlib import Path
import re

import pendulum
from pydantic import BaseModel, SecretStr, root_validator
import smbclient
import smbprotocol

from viadot.config import get_source_credentials
from viadot.exceptions import CredentialError
from viadot.orchestration.prefect.utils import DynamicDateHandler
from viadot.sources.base import Source


class SMBCredentials(BaseModel):
    username: str  # username (e.g username@tenant_name.com)
    password: SecretStr

    @root_validator(pre=True)
    def is_configured(cls, credentials: dict) -> dict:  # noqa: N805
        """Validate that both username and password are provided.

        This method is a Pydantic root validator that checks if both
        username and password fields are present and non-empty.

        Args:
            credentials (dict): A dictionary containing the credential fields.

        Returns:
            dict: The validated credentials dictionary.

        Raises:
            CredentialError: If either username or password is missing.
        """
        username = credentials.get("username")
        password = credentials.get("password")

        if not (username and password):
            msg = "`username`, and `password` credentials are required."
            raise CredentialError(msg)
        return credentials


class SMB(Source):
    def __init__(
        self,
        base_path: str,
        credentials: SMBCredentials | None = None,
        config_key: str | None = None,
        *args,
        **kwargs,
    ):
        """Initialize the SMB with a base path.

        Args:
            base_path (str): The root directory to start scanning from.
            credentials (SMBCredentials): Sharepoint credentials.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials.
        """
        self.base_path = base_path
        raw_creds = credentials or get_source_credentials(config_key) or {}
        validated_creds = SMBCredentials(**raw_creds)
        super().__init__(*args, credentials=validated_creds.dict(), **kwargs)

        normalized_path = re.sub(r"\\+", r"\\", self.base_path)
        parts = normalized_path.lstrip("\\").split("\\")
        server_host_or_ip = parts[0]

        try:
            smbclient.register_session(
                server_host_or_ip,
                username=self.credentials.get("username"),
                password=self.credentials.get("password").get_secret_value(),
            )
            self.logger.info("Connection succesfully established.")
        except smbprotocol.exceptions.LogonFailure:
            self.logger.exception("Authentication failed: credentials invalid.")
            raise
        except smbprotocol.exceptions.PasswordExpired:
            self.logger.exception("Authentication failed: credentials expired.")
            raise
        except Exception:
            self.logger.exception("Connection failed.")
            raise

    def scan_and_store(
        self,
        filename_regex: str | list[str] | None = None,
        extensions: str | list[str] | None = None,
        date_filter: str | tuple[str, str] | None = None,
        dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
        dynamic_date_format: str = "%Y-%m-%d",
        dynamic_date_timezone: str = "UTC",
        prefix_levels_to_add: int = 0,
    ) -> tuple[dict[str, bytes], list[str]]:
        """Scan the directory structure for files and store their contents in memory.

        Args:
            filename_regex (str | list[str] | None, optional): A regular expression
                string or list of regex patterns used to filter file names. If provided,
                only file names matching the pattern(s) will be included.
                Defaults to None.
            extensions (str | list[str] | None): List of file extensions or single
                string to filter by. Defaults to None.
            date_filter (str | tuple[str, str] | None):
                - A single date string (e.g., "2024-03-03").
                - A tuple containing exactly two date strings
                (e.g., ("2024-03-03", "2025-04-04")).
                - None, which returns None.
                Defaults to None.
            dynamic_date_symbols (list[str], optional): Symbols for dynamic date
                handling. Defaults to ["<<", ">>"].
            dynamic_date_format (str, optional): Format used for dynamic date parsing.
                Defaults to "%Y-%m-%d".
            dynamic_date_timezone (str, optional): Timezone used for dynamic date
                processing. Defaults to "UTC".
            prefix_levels_to_add (int, optional): Number of parent folder levels to
                include as a prefix to the filename,counting from the deepest (closest)
                folder upwards. Defaults to 0, meaning no prefix is added.

        Returns:
            tuple[dict[str, bytes], list[str]]:
            - A dictionary mapping file paths to their contents in bytes.
            - A list of file paths that were skipped or failed to be read.
        """
        date_filter_parsed = self._parse_dates(
            date_filter=date_filter,
            dynamic_date_symbols=dynamic_date_symbols,
            dynamic_date_format=dynamic_date_format,
            dynamic_date_timezone=dynamic_date_timezone,
        )

        return self._scan_directory(
            path=self.base_path,
            filename_regex=filename_regex,
            extensions=extensions,
            date_filter_parsed=date_filter_parsed,
            prefix_levels_to_add=prefix_levels_to_add,
        )

    def _parse_dates(
        self,
        date_filter: str | tuple[str, str] | None = None,
        dynamic_date_symbols: list[str] = ["<<", ">>"],  # noqa: B006
        dynamic_date_format: str = "%Y-%m-%d",
        dynamic_date_timezone: str = "UTC",
    ) -> pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None:
        """Parses a date or date range, supporting dynamic date symbols.

        Args:
            date_filter (str | tuple[str, str] | None):
                - A single date string (e.g., "2024-03-03").
                - A tuple containing exactly two date strings, 'start' and 'end' date.
                - None, which applies no date filter.
                Defaults to None.
            dynamic_date_symbols (list[str]): Symbols for dynamic date handling.
                Defaults to ["<<", ">>"].
            dynamic_date_format (str): Format used for dynamic date parsing.
                Defaults to "%Y-%m-%d".
            dynamic_date_timezone (str): Timezone used for dynamic date processing.
                Defaults to "UTC".

        Returns:
            pendulum.Date: If a single date is provided.
            tuple[pendulum.Date, pendulum.Date]: If a date range is provided.
            None: If `date_filter` is None.

        Raises:
            ValueError: If `date_filter` is neither a string nor a tuple of exactly
                two strings.
        """
        if date_filter is None:
            return None

        ddh = DynamicDateHandler(
            dynamic_date_symbols=dynamic_date_symbols,
            dynamic_date_format=dynamic_date_format,
            dynamic_date_timezone=dynamic_date_timezone,
        )

        match date_filter:
            case str():
                return pendulum.parse(ddh.process_dates(date_filter)).date()

            case (start, end) if isinstance(start, str) and isinstance(end, str):
                return (
                    pendulum.parse(ddh.process_dates(start)).date(),
                    pendulum.parse(ddh.process_dates(end)).date(),
                )

            case _:
                msg = (
                    "date_filter must be a string, a tuple of exactly 2 dates, or None."
                )
                raise ValueError(msg)

    def _scan_directory(
        self,
        path: str,
        filename_regex: str | list[str] | None = None,
        extensions: str | list[str] | None = None,
        date_filter_parsed: pendulum.Date
        | tuple[pendulum.Date, pendulum.Date]
        | None = None,
        prefix_levels_to_add: int = 0,
    ) -> tuple[dict[str, bytes], list[str]]:
        """Recursively scans a directory for matching files based on filters.

        It applies 'filename_regex' and 'extensions' filters and can filter files based
        on modification dates - 'date_filter_parsed'.

        It applies filters to both files and directories. Directories are only scanned
        recursively if their modification date matches the given filter.
        This optimization avoids unnecessary traversal of unchanged folders.

        Args:
            path (str): The directory path to scan.
            filename_regex (str | list[str] | None, optional): A regular expression
                string or list of regex patterns used to filter file names. If provided,
                only file names matching the pattern(s) will be included.
                Defaults to None.
            extensions (str | list[str] | None): List of file extensions or single
                string to filter by. Defaults to None.
            date_filter_parsed (
                pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None
            ):
                - A single `pendulum.Date` for exact date filtering.
                - A tuple of two `pendulum.Date` values for date range filtering.
                - None, if no date filter is applied.
                Defaults to None.
            prefix_levels_to_add (int, optional): Number of parent folder levels to
                include as a prefix to the filename,counting from the deepest (closest)
                folder upwards. Defaults to 0, meaning no prefix is added.

        Returns:
            tuple[dict[str, bytes], list[str]]:
            - A dictionary mapping file paths to their contents in bytes.
            - A list of file paths that were skipped or failed to be read.
        """
        found_files = {}
        problematic_entries = []

        entries = self._get_directory_entries(path)
        for entry in entries:
            # Skip temp files
            if entry.name.startswith("~$"):
                problematic_entries.append(entry.name)
                continue

            entry_mod_date_parsed = pendulum.from_timestamp(
                entry.stat().st_mtime
            ).date()
            entry_name = entry.name

            try:
                if entry.is_file() and self._is_matching_file(
                    file_name=entry_name,
                    file_mod_date_parsed=entry_mod_date_parsed,
                    filename_regex=filename_regex,
                    extensions=extensions,
                    date_filter_parsed=date_filter_parsed,
                ):
                    found_files.update(
                        self._get_file_content(entry, prefix_levels_to_add)
                    )

                elif entry.is_dir():
                    date_match = self._is_date_match(
                        entry_mod_date_parsed, date_filter_parsed
                    )

                    if date_match:
                        found_files.update(
                            self._scan_directory(
                                entry.path,
                                filename_regex,
                                extensions,
                                date_filter_parsed,
                                prefix_levels_to_add,
                            )[0]  # Only the matched files dict is used
                        )
            except smbprotocol.exceptions.SMBOSError as e:
                self.logger.warning(f"Entry not found: {e}")
                problematic_entries.append(entry.name)
            except Exception:
                self.logger.exception(f"Error scanning or downloading from {path}.")
                raise

        return found_files, problematic_entries

    def _add_prefix_to_file_name(
        self, file_path: str, prefix_levels_to_add: int = 0
    ) -> str:
        """Generate a new filename by adding parent folder names as prefix.

        Args:
            file_path (str): The full or relative path to the file.
            prefix_levels_to_add (int, optional): Number of parent folder levels to
                include as a prefix to the filename,counting from the deepest (closest)
                folder upwards. Defaults to 0, meaning no prefix is added.

        Returns:
            str: New filename with the specified parent folders as prefix separated by
                underscores.If no prefix levels are specified, returns the original
                filename.
        """
        normalized_path = file_path.replace("\\", "/")
        path = Path(normalized_path)

        parent_parts = [
            p for p in path.parent.parts if p and p not in ("/", "\\", ".", "//", "")
        ]

        # Normalize prefix level count to valid range
        levels = max(0, prefix_levels_to_add)
        levels = min(levels, len(parent_parts))

        prefix_parts_to_add = parent_parts[-levels:] if levels > 0 else []

        prefix_str = "_".join(prefix_parts_to_add)

        return "_".join([prefix_str, path.name]) if prefix_str else path.name

    def _get_file_content(
        self, entry: smbclient._os.SMBDirEntry, prefix_levels_to_add: int = 0
    ) -> dict[str, bytes]:
        """Extracts the content of a file from an SMB directory entry.

        This function takes an SMB directory entry, logs the file path,
        fetches the file's content, and returns a dictionary with the
        file name as the key and its content as the value.

        Args:
            entry (smbclient._os.SMBDirEntry): An SMB directory entry object.
            prefix_levels_to_add (int, optional): Number of parent folder levels to
                include as a prefix to the filename,counting from the deepest (closest)
                folder upwards. Defaults to 0, meaning no prefix is added.

        Returns:
            dict[str, bytes]: A dictionary with a single key-value pair, where the key
                is the file name and the value is the file's content.
        """
        file_path = entry.path

        self.logger.info(f"Found: {file_path}")

        with smbclient.open_file(file_path, mode="rb") as file:
            content = file.read()

        file_name = self._add_prefix_to_file_name(
            file_path=file_path, prefix_levels_to_add=prefix_levels_to_add
        )

        return {file_name: content}

    def _get_directory_entries(self, path: str):
        """Get directory entries using smbclient.

        Args:
            path (str): The directory path to scan.

        Returns:
            Iterator: An iterator of directory entries.
        """
        return smbclient.scandir(path)

    def _is_date_match(
        self,
        file_modification_date: pendulum.Date,
        date_filter_parsed: pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None,
    ) -> bool:
        """Check if the file modification date matches the given date filter.

        Args:
            file_modification_date (pendulum.Date): The modification date of the file.
            date_filter_parsed (
                pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None
            ):
                - A single `pendulum.Date` for exact date filtering.
                - A tuple of two `pendulum.Date` values for date range filtering.
                - None, if no date filter is applied. Defaults to None.

        Returns:
            bool: True if the file_modification_date matches the filter or if no filter
                is applied. False otherwise.
        """
        if date_filter_parsed is None:
            return True

        if isinstance(date_filter_parsed, pendulum.Date):
            return file_modification_date == date_filter_parsed

        if isinstance(date_filter_parsed, tuple):
            start_date, end_date = date_filter_parsed
            return start_date <= file_modification_date <= end_date

        return False

    def _is_matching_file(
        self,
        file_name: str,
        file_mod_date_parsed: pendulum.Date,
        filename_regex: str | list[str] | None = None,
        extensions: str | list[str] | None = None,
        date_filter_parsed: pendulum.Date
        | tuple[pendulum.Date, pendulum.Date]
        | None = None,
    ) -> bool:
        """Check if a file matches the given criteria.

        It verifies whether the file satisfies any combination of:
        - Filename regular expression filtering.
        - Extension-based filtering.
        - Exact date or date range filtering.

        Args:
            file_name (str): The name of the file to evaluate.
            file_mod_date_parsed (pendulum.Date): The parsed modification date of
                the file.
            filename_regex (str | list[str] | None, optional): A regular expression
                string or list of regex patterns used to filter file names. If provided,
                only file names matching the pattern(s) will be included.
                Defaults to None.
            extensions (str | list[str] | None): List of file extensions or single
                string to filter by. It is case-insensitive. Defaults to None.
            date_filter_parsed (
                pendulum.Date | tuple[pendulum.Date, pendulum.Date] | None
            ):
                - A single `pendulum.Date` for exact date filtering.
                - A tuple of two `pendulum.Date` values for date range filtering.
                - None, if no date filter is applied. Defaults to None.

        Returns:
            bool: True if the file matches all criteria or no criteria are provided,
                False otherwise.
        """
        name_lower = file_name.lower()

        # Normalize to lists
        filename_regex_list = (
            [filename_regex] if isinstance(filename_regex, str) else filename_regex
        )
        extension_list = [extensions] if isinstance(extensions, str) else extensions

        matches_extension = not extension_list or any(
            isinstance(ext, str) and name_lower.endswith(ext.lower())
            for ext in extension_list
        )

        matches_filename = (
            True
            if not filename_regex_list
            else any(
                self._safe_regex_match(pattern, name_lower)
                for pattern in filename_regex_list
            )
        )

        if not matches_extension or not matches_filename:
            return False

        if date_filter_parsed:
            return self._is_date_match(file_mod_date_parsed, date_filter_parsed)

        return True

    def _safe_regex_match(self, pattern: str, text: str) -> bool:
        """Evaluate whether a regex pattern matches given text (case-insensitive).

        This method wraps `re.search` with error handling to catch and log invalid regex
        patterns without interrupting execution.

        Args:
            pattern (str): The regular expression pattern to match against.
            text (str): The input string to search within.

        Returns:
            bool: True if the pattern matches the text; False if it does not match
                or if the pattern is invalid.
        """
        try:
            return re.search(pattern, text, re.IGNORECASE) is not None
        except re.error as e:
            self.logger.warning(f"Invalid regex pattern: {pattern} — Error: {e}")
            return False

    def save_files_locally(
        self, file_data: dict[str, bytes], destination_dir: str
    ) -> None:
        """Save stored files from memory to a local directory.

        This function takes a dictionary of file names and their contents,
        and saves each file to the specified destination directory. It creates
        the destination directory if it doesn't exist.

        Args:
        file_data (dict[str, bytes]): A dictionary where keys are file names and values
            are file contents as bytes.
        destination_dir (str): The local directory where files should be saved.

        Raises:
            Exception: If there's an error while saving a file, it's logged but
                not raised.

        Note:
        - If files_to_store is empty, the function logs a message and returns.
        - Each file saving operation is wrapped in a try-except block to handle
        potential errors individually.
        """
        if not file_data:
            self.logger.info("No files to save.")
            return

        Path(destination_dir).mkdir(
            parents=True, exist_ok=True
        )  # Ensure the directory exists

        for file_path, content in file_data.items():
            local_filename = Path(destination_dir) / Path(file_path).name
            try:
                with Path(local_filename).open("wb") as f:
                    f.write(content)
                self.logger.info(f"Saved: {local_filename}")
            except Exception as e:
                self.logger.exception(f"Failed to save {local_filename}: {e}")  # noqa: TRY401



"""'customer_gauge.py'."""

from datetime import datetime
from typing import Any, Literal

import pandas as pd
from pydantic import BaseModel

from viadot.config import get_source_credentials
from viadot.exceptions import APIError
from viadot.sources.base import Source
from viadot.utils import (
    add_viadot_metadata_columns,
    anonymize_df,
    handle_api_response,
    validate,
)


class CustomerGaugeCredentials(BaseModel):
    """Checking for values in Customer Gauge credentials dictionary.

    One key value is held in the Customer Gauge connector:
        - client_id: Unique cliend id number.
        - client_secret: Unique client secret.

    Args:
        BaseModel (pydantic.main.ModelMetaclass): A base class for creating
            Pydantic models.
    """

    client_id: str
    client_secret: str


class CustomerGauge(Source):
    """A class to connect and download data using Customer Gauge API.

    Documentation for each of this API's gateways:
        Responses gateway https://support.customergauge.com/support/solutions/
            articles/5000875861-get-responses
        Non-Responses gateway https://support.customergauge.com/support/solutions/
            articles/5000877703-get-non-responses
    """

    API_URL = "https://api.eu.customergauge.com/v7/rest/sync/"

    def __init__(
        self,
        *args,
        credentials: CustomerGaugeCredentials | None = None,
        config_key: str = "customer_gauge",
        **kwargs,
    ):
        """A class to connect and download data using Customer Gauge API.

        Args:
            credentials (CustomerGaugeCredentials, optional): Customer Gauge
                credentials. Defaults to None.
            config_key (str, optional): The key in the viadot config holding relevant
                credentials. Defaults to "customer_gauge".

        Raises:
            CredentialError: If credentials are not provided in local_config or
                directly as a parameter.
        """
        raw_creds = credentials or get_source_credentials(config_key)
        validated_creds = dict(CustomerGaugeCredentials(**raw_creds))

        super().__init__(*args, credentials=validated_creds, **kwargs)

        self.clean_json = None

    def _get_token(self) -> str:
        """Gets Bearer Token using POST request method.

        Raises:
            APIError: If token is not returned.

        Returns:
            str: Bearer Token value.
        """
        url = "https://auth.EU.customergauge.com/oauth2/token"
        client_id = self.credentials.get("client_id", None)
        client_secret = self.credentials.get("client_secret", None)

        headers = {"Content-Type": "application/x-www-form-urlencoded"}
        body = {
            "grant_type": "client_credentials",
            "client_id": client_id,
            "client_secret": client_secret,
        }
        api_response = handle_api_response(
            url=url, params=body, headers=headers, method="POST"
        )
        token = api_response.json().get("access_token")

        if token is None:
            message = "The token could not be generated. Check your credentials."
            raise APIError(message)

        return token

    def _field_reference_unpacker(
        self,
        json_response: dict[str, Any],
        field: str,
    ) -> dict[str, Any]:
        """Unpack and modify dictionaries within the specified field of a JSON response.

        This function takes a JSON response and a field name. It processes dictionaries
        within the specified field, checking if each dictionary contains
        exactly two items. If a dictionary meets this criteria, it is transformed
        into a new dictionary, where the first key becomes a key, and the second key
        becomes its associated value

        Args:
            json_response (dict[str, Any], optional): JSON response with data.
            field (str): The key (column) of the dictionary to be modified.

        Returns:
            dict[str, Any]: The JSON response with modified nested dictionaries
            within the specified field.

        Raises:
            ValueError: If a dictionary within the specified field doesn't contain
                exactly two items.
        """
        result = {}
        for _, dictionary in enumerate(json_response[field]):
            if isinstance(dictionary, dict) and len(dictionary.items()) == 2:  # noqa: PLR2004
                list_properties = list(dictionary.values())
                result[list_properties[0]] = list_properties[1]
            else:
                message = (
                    "Dictionary within the specified field "
                    "doesn't contain exactly two items."
                )
                raise ValueError(message)
        if result:
            json_response[field] = result

        return json_response

    def _nested_dict_transformer(
        self,
        json_response: dict[str, Any],
        field: str,
    ) -> dict[str, Any]:
        """Modify nested dictionaries within the specified field of a JSON response.

        This function takes a JSON response and a field name. It modifies nested
        dictionaries within the specified field by adding an index and underscore
        to the keys. The modified dictionary is then updated in the JSON response.

        Args:
            json_response (dict[str, Any], optional): JSON response with data.
            field (str): The key (column) of the dictionary to be modified.

        Returns:
            dict[str, Any]: The JSON response with modified nested dictionaries
                within the specified field.
        """
        result = {}
        try:
            for i, dictionary in enumerate(json_response[field], start=1):
                for key, value in dictionary.items():
                    result[f"{i}_{key}"] = value
            if result:
                json_response[field] = result
        except TypeError as type_error:
            self.logger.exception(type_error)  # noqa: TRY401
            raise

        return json_response

    def _get_json_response(
        self,
        url: str | None = None,
        cursor: int | None = None,
        pagesize: int = 1000,
        date_field: str | None = None,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
    ) -> dict[str, Any]:
        """Get JSON that contains data and cursor parameter value.

        Args:
            url (str, optional): URL API direction. Defaults to None.
            cursor (int, optional): Cursor value to navigate to the page.
                Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page,
                max value = 1000. Defaults to 1000. Defaults to 1000.
            date_field (str, optional): Specifies the date type which filter date range.
                Possible options: "date_creation", "date_order", "date_sent" or
                "date_survey_response". Defaults to None.
            start_date (datetime, optional): Defines the period start date in
                yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period end date in
                yyyy-mm-dd format. Defaults to None.

        Raises:
            ValueError: If at least one date argument were provided and the rest is
                missing. Needed all 3 or skip them.
            APIError: If no response from API call.

        Returns:
            dict[str, Any]: JSON with data and cursor parameter value.
        """
        params = {
            "per_page": pagesize,
            "with[]": ["drivers", "tags", "questions", "properties"],
            "cursor": cursor,
        }

        if any([date_field, start_date, end_date]):
            if all([date_field, start_date, end_date]):
                params["period[field]"] = (date_field,)
                params["period[start]"] = (start_date,)
                params["period[end]"] = end_date
            else:
                message = (
                    "Missing date arguments: 'date_field', 'start_date', "
                    "'end_date'. Provide all 3 arguments or skip all of them."
                )
                raise ValueError(message)

        header = {"Authorization": f"Bearer {self._get_token()}"}
        api_response = handle_api_response(url=url, headers=header, params=params)
        response = api_response.json()

        if response is None:
            message = "No response."
            raise APIError(message)

        return response

    def _get_cursor(self, json_response: dict[str, Any] | None = None) -> int:
        """Returns cursor value that is needed to navigate to the next page.

        Args:
            json_response (dict[str, Any], optional): Dictionary with nested structure
                that contains data and cursor parameter value. Defaults to None.

        Raises:
            ValueError: If cursor value not found.

        Returns:
            int: Cursor value.
        """
        try:
            cur = json_response["cursor"]["next"]
        except KeyError as error:
            message = (
                "Provided argument doesn't contain 'cursor' value. "
                "Pass json returned from the endpoint."
            )
            raise ValueError(message) from error

        return cur

    def _get_data(
        self,
        json_response: dict[str, Any] | None = None,
    ) -> list[dict[str, Any]]:
        """Extract the 'data' part of a JSON response as a list of dictionaries.

        Args:
            json_response (Dict[str, Any], optional): JSON object represented as a
                nested dictionary that contains data and cursor parameter value.
                Defaults to None.

        Raises:
            KeyError: If the 'data' key is not present in the provided JSON response.

        Returns:
            list[dict[str, Any]]: A list of dictionaries containing data from the 'data'
                part of the JSON response.
        """
        jsons_list = []
        try:
            jsons_list = json_response["data"]
        except KeyError:
            message = (
                "Provided argument doesn't contain 'data' value. "
                "Pass json returned from the endpoint."
            )
            self.logger.exception(message)
            raise

        return jsons_list

    def _unpack_columns(
        self,
        json_list: list[dict[str, Any]] | None = None,
        unpack_by_field_reference_cols: list[str] | None = None,
        unpack_by_nested_dict_transformer: list[str] | None = None,
    ) -> list[dict[str, Any]]:
        """Function to unpack and modify specific columns in a list of dictionaries.

        It is using one of two methods, chosen by the user.
        If user would like to use field_reference_unpacker, he/she needs to provide
        list of fields as strings in `unpack_by_field_reference_cols`  parameter,
        if user would like to use nested_dict_transformer he/she needs to provide
        list of fields as strings in unpack_by_nested_dict_transformer parameter.

        Args:
            json_list (list[dict[str, Any]): A list of dictionaries containing the data.
            unpack_by_field_reference_cols (list[str]): Columns to unpack and modify
                using `_field_reference_unpacker`. Defaults to None.
            unpack_by_nested_dict_transformer (list[str]): Columns to unpack and
                modify using `_nested_dict_transformer`. Defaults to None.

        Raises:
            ValueError: If 'json_list' is not provided.
            ValueError: If specified columns do not exist in the JSON data.
            ValueError: If columns are mentioned in both
                'unpack_by_field_reference_cols' and 'unpack_by_nested_dict_transformer'

        Returns:
            list[dict[str, Any]]: The updated list of dictionaries after column
                unpacking and modification.
        """
        duplicated_cols = []

        if json_list is None:
            message = "Input 'json_list' is required."
            raise ValueError(message)

        def unpack_columns(columns: list[str], unpack_function: dict[str, Any]):
            json_list_clean = json_list.copy()
            for field in columns:
                if field in json_list_clean[0]:
                    self.logger.info(
                        f"""Unpacking column '{field}' with {unpack_function.__name__}
                        method..."""
                    )
                    try:
                        json_list_clean = list(
                            map(
                                lambda x, f=field: unpack_function(x, f),
                                json_list_clean,
                            )
                        )
                        self.logger.info(
                            f"All elements in '{field}' are unpacked successfully."
                        )
                    except ValueError:
                        self.logger.info(
                            f"No transformation were made in '{field}',"
                            "because didn't contain list of key-value data."
                        )
                else:
                    self.logger.info(f"Column '{field}' not found.")
            return json_list_clean

        if unpack_by_field_reference_cols and unpack_by_nested_dict_transformer:
            duplicated_cols = set(unpack_by_field_reference_cols).intersection(
                set(unpack_by_nested_dict_transformer)
            )
        if duplicated_cols:
            message = (
                f"{duplicated_cols} were mentioned in both"
                "unpack_by_field_reference_cols and unpack_by_nested_dict_transformer."
                "It's not possible to apply two methods to the same field."
            )
            raise ValueError(message)

        if unpack_by_field_reference_cols is not None:
            json_list = unpack_columns(
                columns=unpack_by_field_reference_cols,
                unpack_function=self._field_reference_unpacker,
            )

        if unpack_by_nested_dict_transformer is not None:
            json_list = unpack_columns(
                columns=unpack_by_nested_dict_transformer,
                unpack_function=self._nested_dict_transformer,
            )

        return json_list

    def api_connection(
        self,
        endpoint: Literal["responses", "non-responses"] = "non-responses",
        cursor: int | None = None,
        pagesize: int = 1000,
        date_field: str | None = None,
        start_date: datetime | None = None,
        end_date: datetime | None = None,
        total_load: bool = True,
        unpack_by_field_reference_cols: list[str] | None = None,
        unpack_by_nested_dict_transformer: list[str] | None = None,
    ) -> None:
        """General method to connect to Customer Gauge API and generate the response.

        Args:
            endpoint (Literal["responses", "non-responses"], optional): Indicate which
                endpoint to connect. Defaults to "non-responses.
            cursor (int, optional): Cursor value to navigate to the page.
                Defaults to None.
            pagesize (int, optional): Number of responses (records) returned per page,
                max value = 1000. Defaults to 1000. Defaults to 1000.
            date_field (str, optional): Specifies the date type which filter date range.
                Possible options: "date_creation", "date_order", "date_sent" or
                "date_survey_response". Defaults to None.
            start_date (datetime, optional): Defines the period start date in
                yyyy-mm-dd format. Defaults to None.
            end_date (datetime, optional): Defines the period end date in
                yyyy-mm-dd format. Defaults to None.
            total_load (bool, optional): Indicate whether to download the data to the
                latest. If 'False', only one API call is executed (up to 1000 records).
                Defaults to True.
            unpack_by_field_reference_cols (list[str]): Columns to unpack and modify
                using `_field_reference_unpacker`. Defaults to None.
            unpack_by_nested_dict_transformer (list[str]): Columns to unpack and modify
                using `_nested_dict_transformer`. Defaults to None.
        """
        url = f"{self.API_URL}{endpoint}"
        self.logger.info(f"Starting downloading data from {url} endpoint...")

        total_json = []
        json_data = self._get_json_response(
            url=url,
            cursor=cursor,
            pagesize=pagesize,
            date_field=date_field,
            start_date=start_date,
            end_date=end_date,
        )
        cur = self._get_cursor(json_data)
        jsn = self._get_data(json_data)

        if total_load:
            if cursor is None:
                self.logger.info(
                    f"Downloading all the data from the {endpoint} endpoint."
                    "Process might take a few minutes..."
                )
            else:
                self.logger.info(
                    f"Downloading starting from the {cursor} cursor. "
                    "Process might take a few minutes..."
                )
            while jsn:
                json_data = self._get_json_response(url=url, cursor=cur)
                cur = self._get_cursor(json_data)
                jsn = self._get_data(json_data)
                total_json += jsn

        self.clean_json = self._unpack_columns(
            json_list=total_json,
            unpack_by_field_reference_cols=unpack_by_field_reference_cols,
            unpack_by_nested_dict_transformer=unpack_by_nested_dict_transformer,
        )

    def _flatten_json(
        self, json_response: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Flattens a nested structure of the JSON object.

        Function flattens it into a single-level dictionary. It uses a nested
        `flattify()` function to recursively combine nested keys in the JSON
        object with '_' to create the flattened keys.

        Args:
            json_response (dict[str, Any], optional): JSON object represented as a
                nested dictionary. Defaults to None.

        Raises:
            TypeError: If the 'json_response' not a dictionary.

        Returns:
            dict[str, Any]: The flattened dictionary.
        """
        result = {}

        if not isinstance(json_response, dict):
            message = "Input must be a dictionary."
            raise TypeError(message)

        def flattify(
            field: dict[str, Any], key: str = "", out: dict[str, Any] | None = None
        ):
            if out is None:
                out = result

            if isinstance(field, dict):
                for item in field:
                    flattify(field[item], key + item + "_", out)
            else:
                out[key[:-1]] = field

        flattify(json_response)

        return result

    def _remove_square_brackets(self, df: pd.DataFrame = None) -> pd.DataFrame:
        """Replace square brackets "[]" with an empty string in a pandas DataFrame.

        Args:
            df (pd.DataFrame, optional): Replace square brackets "[]" with an empty
                string in a pandas DataFrame. Defaults to None.

        Returns:
            pd.DataFrame: The modified DataFrame with square brackets replaced by
                an empty string.
        """
        df = df.astype(str)

        return df.map(lambda x: x.strip("[]"))

    def _clean_drivers(self, drivers: str | None = None) -> str:
        """Clean and format the 'drivers' data.

        Args:
            drivers (str, optional): Column name of the data to be cleaned.
                Defaults to None.

        Returns:
            str: A cleaned and formatted string of driver data.
        """
        return (
            drivers.replace("{", "")
            .replace("}", "")
            .replace("'", "")
            .replace("label: ", "")
        )

    @add_viadot_metadata_columns
    def to_df(
        self,
        if_empty: str = "warn",
        validate_df_dict: dict[str, Any] | None = None,
        anonymize: bool = False,
        columns_to_anonymize: list[str] | None = None,
        anonymize_method: Literal["mask", "hash"] = "mask",
        anonymize_value: str = "***",
        date_column: str | None = None,
        days: int | None = None,
    ) -> pd.DataFrame:
        """Generate a Pandas Data Frame with the data in the Response and metadata.

        Args:
            if_empty (str, optional): What to do if a fetch produce no data.
                Defaults to "warn
            validate_df_dict (dict[str, Any], optional): A dictionary with optional list
                of tests to verify the output dataframe. If defined, triggers the
                `validate_df` task from task_utils. Defaults to None.
            anonymize (bool, optional): Indicates if anonymize selected columns.
                Defaults to False.
            columns_to_anonymize (list[str], optional): List of columns to anonymize.
                Defaults to None.
            anonymize_method  (Literal["mask", "hash"], optional): Method of
                anonymizing data. "mask" -> replace the data with "value" arg. "hash" ->
                replace the data with the hash value of an object (using `hash()`
                method). Defaults to "mask".
            anonymize_value (str, optional): Value to replace the data.
                Defaults to "***".
            date_column (str, optional): Name of the date column used to identify rows
                that are older than a specified number of days. Defaults to None.
            days (int, optional): The number of days beyond which we want to anonymize
                the data, e.g. older than 2 years can be: 2*365. Defaults to None.

        Returns:
            pd.Dataframe: The response data as a Pandas Data Frame plus viadot metadata.
        """
        super().to_df(if_empty=if_empty)

        self.logger.info("Inserting data into the DataFrame...")

        data_frame = pd.DataFrame(list(map(self._flatten_json, self.clean_json)))
        data_frame = self._remove_square_brackets(data_frame)

        if "drivers" in list(data_frame.columns):
            data_frame["drivers"] = data_frame["drivers"].apply(self._clean_drivers)
        data_frame.columns = data_frame.columns.str.lower().str.replace(" ", "_")

        if data_frame.empty:
            self._handle_if_empty(
                if_empty=if_empty,
                message="The response does not contain any data.",
            )
        else:
            self.logger.info(
                "Successfully downloaded data from the Customer Gauge API."
            )

        if validate_df_dict is not None:
            validate(df=data_frame, tests=validate_df_dict)

        if anonymize:
            data_frame = anonymize_df(
                df=data_frame,
                columns=columns_to_anonymize,
                method=anonymize_method,
                mask_value=anonymize_value,
                date_column=date_column,
                days=days,
            )
            self.logger.info("Data Frame anonymized")

        return data_frame



import os
from pathlib import Path

from viadot.orchestration.prefect.flows import sharepoint_to_s3
from viadot.sources import S3


S3_BUCKET = os.environ.get("S3_BUCKET")
TEST_SCHEMA = "raw_test"
TEST_TABLE = "test_sharepoint_to_s3"


def test_sharepoint_to_s3(sharepoint_url, sharepoint_config_key):
    file_extension = sharepoint_url.split(".")[-1]
    local_path = "sharepoint_test" + file_extension
    s3_path = f"s3://{S3_BUCKET}/nesso/{TEST_SCHEMA}/{TEST_TABLE}." + file_extension

    sharepoint_to_s3(
        url=sharepoint_url,
        local_path=local_path,
        to_path=s3_path,
        sharepoint_config_key=sharepoint_config_key,
    )

    Path(local_path).unlink()

    s3 = S3()
    file_exists = s3.exists(path=s3_path)

    assert file_exists is True

    s3.rm(paths=[s3_path])



from io import BytesIO
from unittest.mock import MagicMock, patch

from openpyxl import Workbook
import pandas as pd
import pytest
import requests
import sharepy
from sharepy.errors import AuthError

from viadot.exceptions import CredentialError
from viadot.sources import Sharepoint
from viadot.sources.sharepoint import SharepointCredentials


DUMMY_CREDS = {"site": "test", "username": "test2", "password": "test"}
SAMPLE_DF = pd.DataFrame(
    {
        "int_col": [1, 2, 3, 4, 5, None],
        "float_col": [1.1, 2.2, 3.3, 3.0, 5.5, 6.6],
        "str_col": ["a", "b", "c", "d", "e", "f"],
        "nan_col": [None, None, None, None, None, None],
        "mixed_col": [1, "text", None, None, 4.2, "text2"],
    }
)


def create_excel_file():
    """Create a simple Excel file in memory for testing."""
    wb = Workbook()
    ws = wb.active
    ws.title = "Sheet1"
    ws["A1"] = "col_a"
    ws["B1"] = "col_b"
    ws["A2"] = "val1"
    ws["B2"] = "val1"
    ws["A3"] = ""
    ws["B3"] = "val2"
    ws["A4"] = "val2"
    ws["B4"] = "val3"
    ws["A5"] = "NA"
    ws["B5"] = "val4"
    ws["A6"] = "N/A"
    ws["B6"] = "val5"
    ws["A7"] = "#N/A"
    ws["B7"] = "val6"
    bio = BytesIO()
    wb.save(bio)
    bio.seek(0)
    return bio.getvalue()


class SharepointMock(Sharepoint):
    def get_connection(self):
        return sharepy.session.SharePointSession

    def _download_file_stream(self, url: str | None = None, **kwargs):  # noqa: ARG002
        if "nrows" in kwargs:
            msg = "Parameter 'nrows' is not supported."
            raise ValueError(msg)

        return pd.ExcelFile(BytesIO(create_excel_file()))


@pytest.fixture
def sharepoint_mock():
    return SharepointMock(credentials=DUMMY_CREDS)


@pytest.fixture
def sharepoint():
    credentials = {
        "site": "https://example.sharepoint.com",
        "username": "email@example.com",
        "password": "password",
    }
    return Sharepoint(credentials=credentials)


def test_valid_credentials():
    credentials = {
        "site": "tenant.sharepoint.com",
        "username": "user@example.com",
        "password": "password",
    }
    shrp_creds = SharepointCredentials(**credentials)
    assert shrp_creds.site == credentials["site"]
    assert shrp_creds.username == credentials["username"]
    assert shrp_creds.password == credentials["password"]


def test_invalid_authentication():
    credentials = {
        "site": "tenant.sharepoint.com",
        "username": "user@example.com",
        "password": "password",
    }

    s = Sharepoint(credentials=credentials)

    # Patch the sharepy.connect method to simulate an authentication failure
    with patch("sharepy.connect") as mock_connect:
        mock_connect.side_effect = AuthError("Authentication failed")

        with pytest.raises(
            CredentialError,
            match="Could not authenticate to tenant.sharepoint.com with provided credentials.",
        ):
            s.get_connection()


def test_missing_username():
    credentials = {"site": "example.sharepoint.com", "password": "password"}
    with pytest.raises(
        CredentialError,
        match="'site', 'username', and 'password' credentials are required.",
    ):
        SharepointCredentials(**credentials)


def test_sharepoint_default_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx", na_values=Sharepoint.DEFAULT_NA_VALUES
    )

    assert not df.empty
    assert "NA" not in list(df["col_a"])


def test_sharepoint_custom_na(sharepoint_mock):
    df = sharepoint_mock.to_df(
        url="test/file.xlsx",
        na_values=[v for v in Sharepoint.DEFAULT_NA_VALUES if v != "NA"],
    )

    assert not df.empty
    assert "NA" in list(df["col_a"])


def test__get_file_extension(sharepoint_mock):
    url_excel = "https://tenant.sharepoint.com/sites/site/file.xlsx"
    url_dir = "https://tenant.sharepoint.com/sites/site/"
    url_txt = "https://tenant.sharepoint.com/sites/site/file.txt"

    excel_ext = sharepoint_mock._get_file_extension(url=url_excel)
    txt_ext = sharepoint_mock._get_file_extension(url=url_txt)
    dir_ext = sharepoint_mock._get_file_extension(url=url_dir)

    assert excel_ext == ".xlsx"
    assert txt_ext == ".txt"
    assert dir_ext == ""


def test__is_file(sharepoint_mock):
    is_file = sharepoint_mock._is_file(url="https://example.com/file.xlsx")
    assert is_file is True

    is_file = sharepoint_mock._is_file(url="https://example.com/dir")
    assert is_file is False


def test__parse_excel_single_sheet(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")
    expected = pd.DataFrame(
        {
            "col_a": ["val1", "", "val2", "NA", "N/A", "#N/A"],
            "col_b": ["val1", "val2", "val3", "val4", "val5", "val6"],
        }
    )

    assert result_df["col_b"].equals(expected["col_b"])


def test__parse_excel_string_dtypes(sharepoint_mock):
    excel_file = sharepoint_mock._download_file_stream()
    result_df = sharepoint_mock._parse_excel(excel_file, sheet_name="Sheet1")

    for column in result_df.columns:
        assert result_df[column].dtype == object


def test__load_and_parse_not_valid_extension(sharepoint_mock):
    with pytest.raises(ValueError):  # noqa: PT011
        sharepoint_mock._load_and_parse(file_url="https://example.com/file.txt")


def test_scan_sharepoint_folder_valid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/final_folder/"

    # Mock the response from SharePoint
    mock_response = MagicMock()
    mock_response.json.return_value = {
        "d": {
            "results": [
                {"Name": "file1.txt"},
                {"Name": "file2.txt"},
            ]
        }
    }

    # Inject the mock response
    sharepoint_mock.get_connection().get = MagicMock(return_value=mock_response)

    expected_files = [
        "https://company.sharepoint.com/sites/site_name/final_folder/file1.txt",
        "https://company.sharepoint.com/sites/site_name/final_folder/file2.txt",
    ]

    result = sharepoint_mock.scan_sharepoint_folder(url)
    assert result == expected_files


def test_scan_sharepoint_folder_invalid_url(sharepoint_mock):
    url = "https://company.sharepoint.com/folder/sub_folder/final_folder"

    with pytest.raises(ValueError, match="URL does not contain '/sites/' segment."):
        sharepoint_mock.scan_sharepoint_folder(url)


def test_scan_sharepoint_folder_empty_response(sharepoint_mock):
    url = (
        "https://company.sharepoint.com/sites/site_name/folder/sub_folder/final_folder"
    )

    mock_response = MagicMock()
    mock_response.json.return_value = {"d": {"results": []}}

    sharepoint_mock.get_connection().get = MagicMock(return_value=mock_response)

    result = sharepoint_mock.scan_sharepoint_folder(url)
    assert result == []


def test_download_file_stream_unsupported_param(sharepoint_mock):
    url = "https://company.sharepoint.com/sites/site_name/folder/test_file.xlsx"

    with pytest.raises(ValueError, match="Parameter 'nrows' is not supported."):
        sharepoint_mock._download_file_stream(url, nrows=10)


def test_successful_download(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/file.xlsx"
    with patch("sharepy.connect") as mock_connect:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = create_excel_file()
        mock_session = MagicMock()
        mock_session.get.return_value = mock_response
        mock_connect.return_value = mock_session
        df = sharepoint.to_df(url)
        assert not df.empty
        assert len(df) == 6


def test_access_denied(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/restricted_file.xlsx"
    with patch("sharepy.connect") as mock_connect:
        mock_response = MagicMock()
        mock_response.status_code = 403
        mock_response.raise_for_status.side_effect = requests.exceptions.HTTPError(
            response=mock_response
        )
        mock_session = MagicMock()
        mock_session.get.return_value = mock_response
        mock_connect.return_value = mock_session
        with pytest.raises(requests.exceptions.HTTPError):
            sharepoint.to_df(url)


def test_invalid_excel_file(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/file.xlsx"
    with patch("sharepy.connect") as mock_connect:
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = b"not an excel file"
        mock_session = MagicMock()
        mock_session.get.return_value = mock_response
        mock_connect.return_value = mock_session
        with pytest.raises(ValueError, match="Excel file format cannot be determined"):
            sharepoint.to_df(url)


def test_mixed_file_extensions(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/folder"
    with (
        patch("sharepy.connect") as mock_connect,
        patch.object(Sharepoint, "scan_sharepoint_folder") as mock_scan,
    ):
        mock_scan.return_value = [
            url + "/file1.xlsx",
            url + "/file2.txt",
            url + "/file3.pdf",
        ]
        mock_response = MagicMock()
        mock_response.status_code = 200
        mock_response.content = create_excel_file()
        mock_session = MagicMock()
        mock_session.get.return_value = mock_response
        mock_connect.return_value = mock_session

        # Mock _load_and_parse to track which files are processed
        with patch.object(Sharepoint, "_load_and_parse") as mock_load_and_parse:
            mock_load_and_parse.return_value = pd.DataFrame(
                {
                    "col_a": ["val1", "", "val2", "NA", "N/A", "#N/A"],
                    "col_b": ["val1", "val2", "val3", "val4", "val5", "val6"],
                }
            )
            df = sharepoint.to_df(url)

            # Verify that _load_and_parse was only called for .xlsx files
            mock_load_and_parse.assert_called_once_with(
                file_url=url + "/file1.xlsx",
                sheet_name=None,
                na_values=None,
                **{},
            )

        assert not df.empty
        assert len(df) == 6


def test_empty_folder(sharepoint):
    url = "https://example.sharepoint.com/sites/site/Shared%20Documents/folder"
    with (
        patch("sharepy.connect"),
        patch.object(Sharepoint, "scan_sharepoint_folder") as mock_scan,
    ):
        mock_scan.return_value = []
        mock_session = MagicMock()
        mock_connect = patch("sharepy.connect", return_value=mock_session)
        mock_connect.start()
        df = sharepoint.to_df(url)
        mock_connect.stop()
        assert df.empty
